import functools

from django.conf import settings
from django.db import transaction
from django.db.models import Sum

from waldur_core.core import utils as core_utils
from waldur_freeipa import models as freeipa_models

from . import models, tasks, utils


def if_plugin_enabled(f):
    """Calls decorated handler only if plugin is enabled."""
    @functools.wraps(f)
    def wrapped(*args, **kwargs):
        if settings.WALDUR_SLURM['ENABLED']:
            return f(*args, **kwargs)
    return wrapped


@if_plugin_enabled
def process_user_creation(sender, instance, created=False, **kwargs):
    if not created:
        return
    transaction.on_commit(lambda:
                          tasks.add_user.delay(core_utils.serialize_instance(instance)))


@if_plugin_enabled
def process_user_deletion(sender, instance, **kwargs):
    transaction.on_commit(lambda:
                          tasks.delete_user.delay(core_utils.serialize_instance(instance)))


@if_plugin_enabled
def process_role_granted(sender, structure, user, role, **kwargs):
    try:
        freeipa_profile = freeipa_models.Profile.objects.get(user=user)
        serialized_profile = core_utils.serialize_instance(freeipa_profile)
        serialized_structure = core_utils.serialize_instance(structure)
        transaction.on_commit(lambda:
                              tasks.process_role_granted.delay(serialized_profile, serialized_structure))
    except freeipa_models.Profile.DoesNotExist:
        pass


@if_plugin_enabled
def process_role_revoked(sender, structure, user, role, **kwargs):
    try:
        freeipa_profile = freeipa_models.Profile.objects.get(user=user)
        serialized_profile = core_utils.serialize_instance(freeipa_profile)
        serialized_structure = core_utils.serialize_instance(structure)
        transaction.on_commit(lambda:
                              tasks.process_role_revoked.delay(serialized_profile, serialized_structure))
    except freeipa_models.Profile.DoesNotExist:
        pass


@if_plugin_enabled
def update_quotas_on_allocation_usage_update(sender, instance, created=False, **kwargs):
    if created:
        return

    allocation = instance
    if not allocation.usage_changed():
        return

    project = allocation.service_project_link.project
    update_quotas(project, models.Allocation.Permissions.project_path)
    update_quotas(project.customer, models.Allocation.Permissions.customer_path)


def update_quotas(scope, path):
    qs = models.Allocation.objects.filter(**{path: scope}).values(path)
    for quota in utils.FIELD_NAMES:
        qs = qs.annotate(**{'total_%s' % quota: Sum(quota)})
    qs = list(qs)[0]

    for quota in utils.FIELD_NAMES:
        scope.set_quota_usage(utils.MAPPING[quota], qs['total_%s' % quota])



from collections import OrderedDict

from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand
import prettytable
import six

from waldur_core.structure import models

User = get_user_model()

USER_COLUMNS = OrderedDict([
    # (Column name, User fields)
    ('Full name, Civil number', ('full_name', 'civil_number')),
    ('Email, Phone nr.', ('email', 'phone_number')),
    ('Job title', ('job_title',)),
    ('Staff, Support', ('is_staff', 'is_support',)),
])

# in chars
COLUMN_MAX_WIDTH = 25


def format_string_to_column_size(string):
    if len(string) <= COLUMN_MAX_WIDTH:
        return string

    formatted = '\n'.join(string[i:i + COLUMN_MAX_WIDTH] for i in range(0, len(string), COLUMN_MAX_WIDTH))
    if isinstance(formatted, str):
        formatted = six.text_type(formatted, errors='replace')
    return formatted


def to_string(value):
    if isinstance(value, bool):
        return 'Yes' if value else 'No'
    elif isinstance(value, int):
        return str(value)
    elif isinstance(value, six.string_types):
        return format_string_to_column_size(value)
    elif isinstance(value, list):
        strings = [to_string(v) for v in value]
        result = ', '.join(strings)
        if len(result) > COLUMN_MAX_WIDTH:
            return '\n'.join(strings)
        return result

    return format_string_to_column_size(str(value))


class Command(BaseCommand):
    help = "Dumps information about users, their organizations and projects."

    def add_arguments(self, parser):
        parser.add_argument(
            '-o', '--output',
            dest='output', default=None,
            help='Specifies file to which the output is written. The output will be printed to stdout by default.',
        )

    def handle(self, *args, **options):
        # fetch objects
        users = User.objects.all()
        project_roles = models.ProjectPermission.objects.filter(is_active=True)
        customer_roles = models.CustomerPermission.objects.filter(is_active=True)

        # build table
        columns = list(USER_COLUMNS.keys()) + ['Organizations', 'Projects']
        table = prettytable.PrettyTable(columns, hrules=prettytable.ALL)
        for user in users:
            user_customers = to_string(list(customer_roles.filter(user=user)))
            user_projects = to_string(list(project_roles.filter(user=user)))
            row = [to_string([getattr(user, f) for f in fields if getattr(user, f) not in ('', None)])
                   for fields in USER_COLUMNS.values()]
            row += [user_customers, user_projects]
            table.add_row(row)

        # output
        if options['output'] is None:
            self.stdout.write(table.get_string())
            return

        with open(options['output'], 'w') as output_file:
            output_file.write(table.get_string())



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import waldur_core.core.fields
from decimal import Decimal
import django.db.models.deletion
import django.core.validators
import waldur_mastermind.invoices.utils


class Migration(migrations.Migration):

    dependencies = [
        ('support', '0006_offering'),
        ('invoices', '0007_item_remove_price'),
    ]

    operations = [
        migrations.CreateModel(
            name='OfferingItem',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('daily_price', models.DecimalField(default=0, help_text='Price per day.', max_digits=22, decimal_places=7, validators=[django.core.validators.MinValueValidator(Decimal('0'))])),
                ('start', models.DateTimeField(default=waldur_mastermind.invoices.utils.get_current_month_start, help_text='Date and time when item usage has started.')),
                ('end', models.DateTimeField(default=waldur_mastermind.invoices.utils.get_current_month_end, help_text='Date and time when item usage has ended.')),
                ('offering_details', waldur_core.core.fields.JSONField(default={}, help_text='Stores data about offering', blank=True)),
                ('invoice', models.ForeignKey(related_name='+', to='invoices.Invoice')),
                ('offering', models.ForeignKey(related_name='+', on_delete=django.db.models.deletion.SET_NULL, to='support.Offering', null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.16 on 2019-01-17 13:53
from __future__ import unicode_literals

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0050_add_component_limit'),
    ]

    operations = [
        migrations.AddField(
            model_name='resource',
            name='backend_metadata',
            field=django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=dict),
        ),
    ]



import logging

from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.core.management.base import CommandError

from waldur_core.core.utils import DryRunCommand
from waldur_core.structure import models as structure_models
from waldur_core.structure.models import Customer
from waldur_mastermind.marketplace import models as marketplace_models, plugins
from waldur_mastermind.marketplace.utils import format_list
from waldur_mastermind.marketplace_slurm import PLUGIN_NAME
from waldur_slurm import apps as slurm_apps
from waldur_mastermind.slurm_invoices import models as slurm_invoices_models

logger = logging.getLogger(__name__)


class ImportSLURMException(Exception):
    pass


def import_slurm_service_settings(default_customer, dry_run=False):
    """
    Import SLURM service settings as marketplace offerings.
    """
    service_type = slurm_apps.SlurmConfig.service_name
    category = marketplace_models.Category.objects.get(
        uuid=settings.WALDUR_MARKETPLACE_SLURM['CATEGORY_UUID']
    )

    slurm_offerings = marketplace_models.Offering.objects.filter(type=PLUGIN_NAME)
    front_settings = set(slurm_offerings.exclude(object_id=None).values_list('object_id', flat=True))

    back_settings = structure_models.ServiceSettings.objects.filter(type=service_type,
                                                                    state=structure_models.ServiceSettings.States.OK)
    missing_settings = back_settings.exclude(id__in=front_settings)

    if dry_run:
        logger.warning('SLURM service settings would be imported to marketplace. '
                       'IDs: %s.' % format_list(missing_settings))
        return missing_settings.count()

    for service_settings in missing_settings:
        offering = marketplace_models.Offering.objects.create(
            scope=service_settings,
            type=PLUGIN_NAME,
            name=service_settings.name,
            geolocations=service_settings.geolocations,
            customer=service_settings.customer or default_customer,
            category=category,
            shared=service_settings.shared,
            state=marketplace_models.Offering.States.ACTIVE,
        )

        components = plugins.manager.get_components(PLUGIN_NAME)

        for component_data in components:
            marketplace_models.OfferingComponent.objects.create(
                offering=offering,
                **component_data._asdict()
            )

        try:
            slurm_package = slurm_invoices_models.SlurmPackage.objects. \
                get(service_settings=service_settings)
            plan = marketplace_models.Plan.objects.create(
                offering=offering,
                scope=slurm_package,
                name=slurm_package.name)

            marketplace_models.PlanComponent.objects.create(
                plan=plan,
                component=offering.components.filter(type='cpu').get(),
                price=slurm_package.cpu_price)
            marketplace_models.PlanComponent.objects.create(
                plan=plan,
                component=offering.components.filter(type='gpu').get(),
                price=slurm_package.gpu_price)
            marketplace_models.PlanComponent.objects.create(
                plan=plan,
                component=offering.components.filter(type='ram').get(),
                price=slurm_package.ram_price)
        except slurm_invoices_models.SlurmPackage.DoesNotExist:
            logger.warning('Plan has not been created. Because SlurmPackage is not found. '
                           'Service settings UUID: %s.' % service_settings.uuid.hex)

    return missing_settings.count()


class Command(DryRunCommand):
    help = """Import SLURM service settings as marketplace offerings."""

    def add_arguments(self, parser):
        super(Command, self).add_arguments(parser)
        parser.add_argument('--customer', dest='customer_uuid', required=True,
                            help='Default customer argument is used for shared service setting.')

    def handle(self, customer_uuid, dry_run, *args, **options):
        try:
            customer = Customer.objects.get(uuid=customer_uuid)
        except ObjectDoesNotExist:
            raise CommandError('A customer is not found.')

        try:
            offerings_counter = import_slurm_service_settings(customer, dry_run)
            self.stdout.write(self.style.SUCCESS('%s offerings have been created.' % offerings_counter))
        except ImportSLURMException as e:
            raise CommandError(e.message)
        except marketplace_models.Category.DoesNotExist:
            raise CommandError('Please ensure that WALDUR_MARKETPLACE_SLURM.CATEGORY_UUID '
                               'setting has valid value.')



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('invoices', '0004_paymentdetails_uuid'),
    ]

    operations = [
        migrations.AddField(
            model_name='paymentdetails',
            name='type',
            field=models.CharField(max_length=150, blank=True),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.18 on 2019-03-25 16:19
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0066_cartitem_cost'),
    ]

    operations = [
        migrations.AddField(
            model_name='offeringcomponent',
            name='disable_quotas',
            field=models.BooleanField(default=False, help_text='Do not allow user to specify quotas when offering is provisioned.'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.13 on 2017-04-10 16:46
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('openstack_tenant', '0025_copy_certifications_from_existing_settings'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='backup',
            name='start_time',
        ),
        migrations.RemoveField(
            model_name='backupschedule',
            name='start_time',
        ),
        migrations.RemoveField(
            model_name='snapshot',
            name='start_time',
        ),
        migrations.RemoveField(
            model_name='snapshotschedule',
            name='start_time',
        ),
        migrations.RemoveField(
            model_name='volume',
            name='start_time',
        ),
    ]



from celery import chain

from waldur_core.core import executors, tasks, utils

from . import models
from .tasks import SMSTask, UpdateSettingsCredentials


class HostCreateExecutor(executors.CreateExecutor):

    @classmethod
    def get_task_signature(cls, host, serialized_host, **kwargs):
        return tasks.BackendMethodTask().si(
            serialized_host, 'create_host', state_transition='begin_creating')


class HostUpdateExecutor(executors.UpdateExecutor):

    @classmethod
    def get_task_signature(cls, host, serialized_host, **kwargs):
        return tasks.BackendMethodTask().si(
            serialized_host, 'update_host', state_transition='begin_updating')


class HostDeleteExecutor(executors.DeleteExecutor):

    @classmethod
    def get_task_signature(cls, host, serialized_host, **kwargs):
        if host.backend_id:
            return tasks.BackendMethodTask().si(
                serialized_host, 'delete_host', state_transition='begin_deleting')
        else:
            return tasks.StateTransitionTask().si(serialized_host, state_transition='begin_deleting')


class HostPullExecutor(executors.BaseExecutor):

    @classmethod
    def get_task_signature(cls, host, serialized_host, **kwargs):
        return tasks.BackendMethodTask().si(serialized_host, 'pull_host')


class ITServiceCreateExecutor(executors.CreateExecutor):

    @classmethod
    def get_task_signature(cls, itservice, serialized_itservice, **kwargs):
        return tasks.BackendMethodTask().si(
            serialized_itservice, 'create_itservice', state_transition='begin_creating')


class ITServiceDeleteExecutor(executors.DeleteExecutor):

    @classmethod
    def get_task_signature(cls, itservice, serialized_itservice, **kwargs):
        if itservice.backend_id:
            return tasks.BackendMethodTask().si(
                serialized_itservice, 'delete_itservice', state_transition='begin_deleting')
        else:
            return tasks.StateTransitionTask().si(serialized_itservice, state_transition='begin_deleting')


class UserCreateExecutor(executors.CreateExecutor):

    @classmethod
    def get_task_signature(cls, user, serialized_user, **kwargs):
        creation_tasks = [tasks.BackendMethodTask().si(
            serialized_user, 'create_user', state_transition='begin_creating')]
        # send SMS if user has phone number
        if user.phone:
            serialized_settings = utils.serialize_instance(user.settings)
            message = 'Zabbix "%s" password: %s' % (user.settings.name, user.password)
            creation_tasks.append(SMSTask().si(serialized_settings, message, user.phone))
        return chain(*creation_tasks)


class UserUpdateExecutor(executors.UpdateExecutor):

    @classmethod
    def get_task_signature(cls, user, serialized_user, **kwargs):
        update_tasks = [tasks.BackendMethodTask().si(
            serialized_user, 'update_user', state_transition='begin_updating')]
        # send SMS if user password has been updated
        if 'password' in kwargs.get('updated_fields', []) and user.phone:
            serialized_settings = utils.serialize_instance(user.settings)
            message = 'Zabbix "%s" password: %s' % (user.settings.name, user.password)
            update_tasks.append(SMSTask().si(serialized_settings, message, user.phone))
        return chain(*update_tasks)


class UserDeleteExecutor(executors.DeleteExecutor):

    @classmethod
    def get_task_signature(cls, user, serialized_user, **kwargs):
        if user.backend_id:
            return tasks.BackendMethodTask().si(
                serialized_user, 'delete_user', state_transition='begin_deleting')
        else:
            return tasks.StateTransitionTask().si(serialized_user, state_transition='begin_deleting')


class ServiceSettingsPasswordResetExecutor(executors.ActionExecutor):
    """ Reset user password and update service settings options. """

    @classmethod
    def get_task_signature(cls, service_settings, serialized_service_settings, **kwargs):
        user = models.User.objects.get(settings=service_settings, alias=service_settings.username)
        user.password = kwargs.pop('password')
        user.schedule_updating()
        user.save()
        serialized_user = utils.serialize_instance(user)
        _tasks = [
            tasks.StateTransitionTask().si(serialized_service_settings, state_transition='begin_updating'),
            tasks.BackendMethodTask().si(serialized_user, 'update_user', state_transition='begin_updating'),
            UpdateSettingsCredentials().si(serialized_service_settings, serialized_user),
            tasks.StateTransitionTask().si(serialized_user, state_transition='set_ok'),
        ]
        if user.phone:
            message = 'Zabbix "%s" password: %s' % (user.settings.name, user.password)
            _tasks.append(SMSTask().si(serialized_service_settings, message, user.phone))
        return chain(*_tasks)



from __future__ import unicode_literals

from django.contrib import admin
from django.core.urlresolvers import resolve
from django.forms.models import ModelForm
from django.shortcuts import redirect
from django.urls import reverse
from django.utils.html import format_html
from django.utils.translation import ugettext_lazy as _, ungettext
from waldur_core.core.admin import JsonWidget

from waldur_core.core import admin as core_admin
from waldur_core.core.admin import format_json_field
from waldur_core.core.admin_filters import RelatedOnlyDropdownFilter
from waldur_core.structure.models import ServiceSettings, SharedServiceSettings, PrivateServiceSettings

from . import models, tasks


class ServiceProviderAdmin(admin.ModelAdmin):
    list_display = ('uuid', 'customer', 'created')


class AttributeOptionInline(admin.TabularInline):
    model = models.AttributeOption


class AttributeAdmin(admin.ModelAdmin):
    inlines = [AttributeOptionInline]
    list_display = ('title', 'get_category', 'section', 'type', 'key', 'required',)
    list_filter = ('section',)
    ordering = ('section', 'title')

    def get_category(self, obj):
        return obj.section.category

    get_category.short_description = _('Category')
    get_category.admin_order_field = 'section__category__title'


class AttributeInline(admin.TabularInline):
    model = models.Attribute


class SectionAdmin(admin.ModelAdmin):
    inlines = [AttributeInline]
    list_display = ('title', 'category', 'key')


class SectionInline(admin.TabularInline):
    model = models.Section


class CategoryColumnInline(admin.TabularInline):
    model = models.CategoryColumn
    list_display = ('index', 'title', 'attribute', 'widget')


class CategoryComponentInline(admin.TabularInline):
    model = models.CategoryComponent


class CategoryAdmin(admin.ModelAdmin):
    list_display = ('title', 'uuid',)
    inlines = [SectionInline, CategoryColumnInline, CategoryComponentInline]


class ScreenshotsInline(admin.StackedInline):
    model = models.Screenshot
    fields = ('name', 'description', 'image')


class PlansInline(admin.StackedInline):
    model = models.Plan
    fields = ('name', 'description', 'unit_price', 'unit',
              'product_code', 'article_code', 'archived', 'max_amount')


class ConnectedResourceMixin(object):
    """
    Protects object from modification if there are connected resources.
    """

    protected_fields = ()

    def get_readonly_fields(self, request, obj=None):
        fields = super(ConnectedResourceMixin, self).get_readonly_fields(request, obj)
        if obj and obj.has_connected_resources:
            return fields + self.protected_fields
        else:
            return fields

    def has_delete_permission(self, request, obj=None):
        if obj and obj.has_connected_resources:
            return False
        return True


class ParentInlineMixin(object):
    def get_parent_object_from_request(self, request):
        """
        Returns the parent object from the request or None.

        Note that this only works for Inlines, because the `parent_model`
        is not available in the regular admin.ModelAdmin as an attribute.
        """
        resolved = resolve(request.path_info)
        if resolved.args:
            return self.parent_model.objects.get(pk=resolved.args[0])
        return None


class PlanComponentInline(ConnectedResourceMixin,
                          ParentInlineMixin,
                          admin.TabularInline):
    model = models.PlanComponent
    protected_fields = ('component', 'amount', 'price')

    def has_add_permission(self, request, obj=None):
        plan = self.get_parent_object_from_request(request)
        if plan and plan.has_connected_resources:
            return False
        else:
            return True

    def get_extra(self, request, obj=None, **kwargs):
        plan = self.get_parent_object_from_request(request)
        if plan and plan.has_connected_resources:
            return 0
        else:
            return super(PlanComponentInline, self).get_extra(request, obj, **kwargs)


class PlanAdmin(ConnectedResourceMixin, admin.ModelAdmin):
    list_display = ('name', 'offering', 'archived', 'unit', 'unit_price')
    list_filter = ('offering', 'archived')
    search_fields = ('name', 'offering__name')
    inlines = [PlanComponentInline]
    protected_fields = ('unit', 'unit_price', 'product_code', 'article_code')
    readonly_fields = ('scope_link', 'backend_id')
    fields = (
        'name', 'description', 'unit', 'unit_price',
        'product_code', 'article_code', 'max_amount',
        'archived',
    ) + readonly_fields

    def scope_link(self, obj):
        return get_admin_link_for_scope(obj.scope)

    scope_link.short_description = 'Scope'


class OfferingAdminForm(ModelForm):
    class Meta:
        widgets = {
            'attributes': JsonWidget(),
            'geolocations': JsonWidget(),
            'options': JsonWidget(),
        }


class OfferingComponentInline(admin.StackedInline):
    model = models.OfferingComponent


def get_admin_url_for_scope(scope):
    if isinstance(scope, ServiceSettings):
        model = scope.shared and SharedServiceSettings or PrivateServiceSettings
    else:
        model = scope
    return reverse('admin:%s_%s_change' % (scope._meta.app_label, model._meta.model_name), args=[scope.id])


def get_admin_link_for_scope(scope):
    return format_html('<a href="{}">{}</a>', get_admin_url_for_scope(scope), scope)


class OfferingAdmin(admin.ModelAdmin):
    form = OfferingAdminForm
    inlines = [ScreenshotsInline, PlansInline, OfferingComponentInline]
    list_display = ('name', 'customer', 'state', 'category', 'billable')
    list_filter = ('state', 'shared', 'billable', ('category', RelatedOnlyDropdownFilter),)
    search_fields = ('name', 'uuid')
    fields = ('state', 'customer', 'category', 'name', 'native_name',
              'description', 'native_description', 'full_description',
              'rating', 'thumbnail', 'attributes', 'options', 'geolocations',
              'shared', 'billable', 'allowed_customers', 'type', 'scope_link', 'vendor_details')
    readonly_fields = ('rating', 'scope_link')

    def scope_link(self, obj):
        if obj.scope:
            return format_html('<a href="{}">{}</a>', get_admin_url_for_scope(obj.scope), obj.scope)

    actions = ['activate']

    def activate(self, request, queryset):
        valid_states = [models.Offering.States.DRAFT, models.Offering.States.PAUSED]
        valid_offerings = queryset.filter(state__in=valid_states)
        count = valid_offerings.count()

        for offering in valid_offerings:
            offering.activate()
            offering.save()

        message = ungettext(
            'One offering has been activated.',
            '%(count)d offerings have been activated.',
            count
        )
        message = message % {'count': count}

        self.message_user(request, message)

    activate.short_description = _('Activate offerings')


class OrderItemInline(admin.TabularInline):
    model = models.OrderItem
    fields = ('offering', 'state', 'attributes', 'cost', 'plan')
    readonly_fields = fields


class OrderAdmin(core_admin.ExtraActionsMixin, admin.ModelAdmin):
    list_display = ('uuid', 'project', 'created', 'created_by', 'state', 'total_cost')
    list_filter = ('state', 'created')
    ordering = ('-created',)
    inlines = [OrderItemInline]

    def get_extra_actions(self):
        return [
            self.create_pdf_for_all,
        ]

    def create_pdf_for_all(self, request):
        tasks.create_pdf_for_all.delay()
        message = _('PDF creation has been scheduled')
        self.message_user(request, message)
        return redirect(reverse('admin:marketplace_order_changelist'))

    create_pdf_for_all.name = _('Create PDF for all orders')


class ResourceForm(ModelForm):
    def __init__(self, *args, **kwargs):
        super(ResourceForm, self).__init__(*args, **kwargs)
        if self.instance and hasattr(self.instance, 'offering'):
            # Filter marketplace resource plans by offering
            self.fields['plan'].queryset = self.fields['plan'].queryset.filter(
                offering=self.instance.offering)


class ResourceAdmin(admin.ModelAdmin):
    form = ResourceForm
    list_display = ('name', 'project', 'state', 'category', 'created')
    list_filter = (
        'state',
        ('project', RelatedOnlyDropdownFilter),
        ('offering', RelatedOnlyDropdownFilter),
    )
    readonly_fields = ('state', 'scope_link', 'project_link', 'offering_link',
                                'plan_link', 'formatted_attributes', 'formatted_limits')
    fields = readonly_fields + ('plan',)
    search_fields = ('name', 'uuid')

    def category(self, obj):
        return obj.offering.category

    def scope_link(self, obj):
        return get_admin_link_for_scope(obj.scope)

    scope_link.short_description = 'Scope'

    def project_link(self, obj):
        return get_admin_link_for_scope(obj.project)

    project_link.short_description = 'Project'

    def offering_link(self, obj):
        return get_admin_link_for_scope(obj.offering)

    offering_link.short_description = 'Offering'

    def plan_link(self, obj):
        return get_admin_link_for_scope(obj.plan)

    plan_link.short_description = 'Plan'

    def formatted_attributes(self, obj):
        return format_json_field(obj.attributes)

    formatted_attributes.allow_tags = True
    formatted_attributes.short_description = 'Attributes'

    def formatted_limits(self, obj):
        return format_json_field(obj.limits)

    formatted_limits.allow_tags = True
    formatted_limits.short_description = 'Limits'


admin.site.register(models.ServiceProvider, ServiceProviderAdmin)
admin.site.register(models.Category, CategoryAdmin)
admin.site.register(models.Offering, OfferingAdmin)
admin.site.register(models.Section, SectionAdmin)
admin.site.register(models.Attribute, AttributeAdmin)
admin.site.register(models.Screenshot)
admin.site.register(models.Order, OrderAdmin)
admin.site.register(models.Plan, PlanAdmin)
admin.site.register(models.Resource, ResourceAdmin)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from uuid import uuid4

from django.db import migrations


def create_quotas(apps, schema_editor):
    Project = apps.get_model('structure', 'Project')
    Customer = apps.get_model('structure', 'Customer')
    ProjectGroup = apps.get_model('structure', 'ProjectGroup')
    Quota = apps.get_model('quotas', 'Quota')

    # We can not use model constants in migrations because they can be changed in future
    quota_name_map = {
        Project: 'nc_global_project_count',
        Customer: 'nc_global_customer_count',
        ProjectGroup: 'nc_global_project_group_count',
    }

    for model in [Project, Customer, ProjectGroup]:
        name = quota_name_map[model]
        usage = model.objects.count()
        if not Quota.objects.filter(name=name, object_id__isnull=True).exists():
            Quota.objects.create(uuid=uuid4().hex, name=name, usage=usage)
        else:
            Quota.objects.filter(name=name, object_id__isnull=True).update(usage=usage)


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0001_squashed_0021_balancehistory'),
        ('quotas', '0002_make_quota_scope_nullable'),
    ]

    operations = [
        migrations.RunPython(create_quotas),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.12 on 2017-03-17 10:50
from __future__ import unicode_literals

from django.db import migrations, models

import waldur_core.core.validators


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0044_terms_of_services_url'),
    ]

    operations = [
        migrations.AddField(
            model_name='project',
            name='certifications',
            field=models.ManyToManyField(blank=True, related_name='projects', to='structure.ServiceCertification'),
        ),
        migrations.AlterField(
            model_name='servicecertification',
            name='name',
            field=models.CharField(max_length=150, unique=True, validators=[waldur_core.core.validators.validate_name], verbose_name='name'),
        ),
    ]



from . import views


def register_in(router):
    router.register(r'azure', views.AzureServiceViewSet, base_name='azure')
    router.register(r'azure-images', views.ImageViewSet, base_name='azure-image')
    router.register(r'azure-sizes', views.SizeViewSet, base_name='azure-size')
    router.register(r'azure-locations', views.LocationViewSet, base_name='azure-location')
    router.register(r'azure-resource-groups', views.ResourceGroupViewSet, base_name='azure-resource-group')
    router.register(r'azure-virtualmachines', views.VirtualMachineViewSet, base_name='azure-virtualmachine')
    router.register(r'azure-public-ips', views.PublicIPViewSet, base_name='azure-public-ip')
    router.register(r'azure-sql-servers', views.SQLServerViewSet, base_name='azure-sql-server')
    router.register(r'azure-sql-databases', views.SQLDatabaseViewSet, base_name='azure-sql-database')
    router.register(r'azure-service-project-link',
                    views.AzureServiceProjectLinkViewSet, base_name='azure-spl')



from waldur_core.core import WaldurExtension


class AnsibleCommonExtension(WaldurExtension):
    class Settings:
        WALDUR_ANSIBLE_COMMON = {
            'PLAYBOOK_EXECUTION_COMMAND': 'ansible-playbook',
            'PLAYBOOK_ARGUMENTS': ['--verbose'],
            'API_URL': 'https://waldur.example.com/api/',
            'PRIVATE_KEY_PATH': '/etc/waldur/id_rsa',
            'PUBLIC_KEY_UUID': 'Corresponding public key should be stored in the database. Specify here its UUID.',
            'ANSIBLE_REQUEST_TIMEOUT': 3600,
            'ANSIBLE_LIBRARY': '/usr/share/ansible-waldur/',
            'REMOTE_VM_SSH_PORT': '22',
        }

    @staticmethod
    def django_app():
        return 'waldur_ansible.common'

    @staticmethod
    def is_assembly():
        return True

    @staticmethod
    def rest_urls():
        from .urls import register_in
        return register_in

    @staticmethod
    def get_public_settings():
        return ['ANSIBLE_REQUEST_TIMEOUT', 'PUBLIC_KEY_UUID']



from waldur_core.core import WaldurExtension


class CostPlanningExtension(WaldurExtension):
    class Settings:
        WALDUR_COST_PLANNING = {
            'currency': 'USD',
        }

    @staticmethod
    def django_app():
        return 'waldur_cost_planning'

    @staticmethod
    def rest_urls():
        from .urls import register_in
        return register_in



from __future__ import unicode_literals

from django.contrib.contenttypes.fields import GenericForeignKey, GenericRelation
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils.translation import ugettext_lazy as _

from waldur_core.core.models import NameMixin
from waldur_core.monitoring.managers import ResourceSlaManager, ResourceItemManager, ResourceSlaStateTransitionManager


class ScopeMixin(models.Model):
    content_type = models.ForeignKey(ContentType)
    object_id = models.PositiveIntegerField()
    scope = GenericForeignKey('content_type', 'object_id')

    class Meta:
        abstract = True


class ResourceItem(NameMixin, ScopeMixin):
    value = models.FloatField()
    objects = ResourceItemManager()

    class Meta:
        unique_together = ('name', 'content_type', 'object_id')


class ResourceSla(ScopeMixin):
    period = models.CharField(max_length=10)
    value = models.DecimalField(max_digits=11, decimal_places=4, null=True, blank=True)
    agreed_value = models.DecimalField(max_digits=11, decimal_places=4, null=True, blank=True)
    objects = ResourceSlaManager()

    class Meta:
        unique_together = ('period', 'content_type', 'object_id')


class ResourceSlaStateTransition(ScopeMixin):
    period = models.CharField(max_length=10)
    timestamp = models.IntegerField()
    state = models.BooleanField(default=False, help_text=_('If state is True resource became available'))
    objects = ResourceSlaStateTransitionManager()

    class Meta:
        unique_together = ('timestamp', 'period', 'content_type', 'object_id')


class MonitoringModelMixin(models.Model):
    class Meta:
        abstract = True

    sla_items = GenericRelation('monitoring.ResourceSla')
    monitoring_items = GenericRelation('monitoring.ResourceItem')
    state_items = GenericRelation('monitoring.ResourceSlaStateTransition')



from __future__ import unicode_literals

from django.http import Http404
from rest_framework import viewsets, mixins, response, decorators

from waldur_core.core import mixins as core_mixins
from waldur_core.core.views import validate_authentication_method

from . import models, serializers, executors


validate_valimo = validate_authentication_method('VALIMO')


class AuthResultViewSet(core_mixins.CreateExecutorMixin,
                        mixins.CreateModelMixin,
                        viewsets.GenericViewSet):
    queryset = models.AuthResult.objects.all()
    serializer_class = serializers.AuthResultSerializer
    permission_classes = ()
    lookup_field = 'uuid'
    create_executor = executors.AuthExecutor

    @validate_valimo
    def create(self, request, *args, **kwargs):
        """
        To start PKI login process - issue post request with users phone.
        Example of a valid request:

        .. code-block:: http

            POST /api/auth-valimo/ HTTP/1.1
            Content-Type: application/json
            Accept: application/json
            Host: example.com

            {
                "phone": "1234567890",
            }
        """
        return super(AuthResultViewSet, self).create(request, *args, **kwargs)

    @validate_valimo
    @decorators.list_route(methods=['POST'])
    def result(self, request, *args, **kwargs):
        """
        To get PKI login status and details - issue post request against /api/auth-valimo/result/
        with uuid in parameters.

        Possible states:
         - Scheduled - login process is scheduled
         - Processing - login is in progress
         - OK - login was successful. Response will contain token.
         - Canceled - login was canceled by user or timed out. Field details will contain additional info.
         - Erred - unexpected exception happened during login process.
        Example of response:

        .. code-block:: javascript

            {
                "uuid": "e42473f39c844333a80107e139a4dd06",
                "token": null,
                "message": "1234",
                "state": "Canceled",
                "error_message": "",
                "details": "User cancel."
            }
        """
        try:
            auth_result = models.AuthResult.objects.get(uuid=request.data.get('uuid'))
        except models.AuthResult.DoesNotExist:
            raise Http404
        serializer = self.get_serializer(auth_result)
        return response.Response(serializer.data)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-06-21 21:33
from __future__ import unicode_literals

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('invoices', '0013_set_projects_for_invoice_items'),
    ]

    operations = [
        migrations.AddField(
            model_name='paymentdetails',
            name='accounting_start_date',
            field=models.DateTimeField('Start date of accounting', default=django.utils.timezone.now),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-04-25 11:43
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0078_fix_plan_component_amount'),
    ]

    operations = [
        migrations.AddField(
            model_name='orderitem',
            name='old_plan',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='marketplace.Plan'),
        ),
    ]



from __future__ import unicode_literals

import logging
from django.contrib.contenttypes.models import ContentType

from waldur_mastermind.invoices import models as invoice_models
from waldur_mastermind.invoices import registrators
from waldur_mastermind.marketplace import models as marketplace_models
from waldur_mastermind.marketplace import utils as marketplace_utils
from waldur_mastermind.support import models as support_models

from . import utils

logger = logging.getLogger(__name__)


class OfferingRegistrator(registrators.BaseRegistrator):
    def get_sources(self, customer):
        return support_models.Offering.objects.filter(
            project__customer=customer,
            state=support_models.Offering.States.OK,
        ).distinct()

    def get_customer(self, source):
        return source.project.customer

    def _find_item(self, source, now):
        offering = source
        result = utils.get_offering_items().filter(
            object_id=offering.id,
            invoice__customer=offering.project.customer,
            invoice__state=invoice_models.Invoice.States.PENDING,
            invoice__year=now.year,
            invoice__month=now.month,
        )
        return list(result)

    def _create_item(self, source, invoice, start, end):
        offering = source

        try:
            resource = marketplace_models.Resource.objects.get(scope=offering)
            plan = resource.plan

            if not plan:
                logger.warning('Skipping support invoice creation because '
                               'billing is not enabled for offering. '
                               'Offering ID: %s', offering.id)
                return

            create, switch = utils.resource_created_or_switched_plan(resource, start)

            fixed = marketplace_models.OfferingComponent.BillingTypes.FIXED
            one_time_type = marketplace_models.OfferingComponent.BillingTypes.ONE_TIME
            switch_type = marketplace_models.OfferingComponent.BillingTypes.ON_PLAN_SWITCH

            for plan_component in plan.components.all():
                offering_component = plan_component.component

                if (offering_component.billing_type == fixed or
                        (offering_component.billing_type == one_time_type and create) or
                        (offering_component.billing_type == switch_type and (switch or create))):
                    details = self.get_component_details(offering, plan_component)
                    invoice_models.GenericInvoiceItem.objects.create(
                        content_type=ContentType.objects.get_for_model(offering),
                        object_id=offering.id,
                        project=offering.project,
                        invoice=invoice,
                        start=start,
                        end=end,
                        details=details,
                        unit_price=plan_component.price,
                        unit=plan.unit,
                        product_code=offering_component.product_code or plan.product_code,
                        article_code=offering_component.article_code or plan.article_code,
                    )

        except marketplace_models.Resource.DoesNotExist:
            # If an offering isn't request based support offering
            return invoice_models.GenericInvoiceItem.objects.create(
                content_type=ContentType.objects.get_for_model(offering),
                object_id=offering.id,
                project=offering.project,
                invoice=invoice,
                start=start,
                end=end,
                details=self.get_details(offering),
                unit_price=offering.unit_price,
                unit=offering.unit,
                product_code=offering.product_code,
                article_code=offering.article_code
            )

    def get_details(self, source):
        offering = source
        details = {
            'name': self.get_name(offering),
            'offering_type': offering.type,
            'offering_name': offering.name,
            'offering_uuid': offering.uuid.hex,
            'plan_name': offering.plan.name if offering.plan else ''
        }
        service_provider_info = marketplace_utils.get_service_provider_info(source)
        details.update(service_provider_info)
        return details

    def get_name(self, offering):
        if offering.plan:
            return '%s (%s / %s)' % (offering.name, offering.type, offering.plan.name)
        else:
            return '%s (%s)' % (offering.name, offering.type)

    def terminate(self, source, now=None):
        super(OfferingRegistrator, self).terminate(source, now)
        offering = source

        if not utils.is_request_based(offering):
            utils.get_offering_items().filter(object_id=offering.id).update(object_id=None)

    def get_component_details(self, offering, plan_component):
        details = self.get_details(offering)
        details.update({
            'plan_component_id': plan_component.id,
            'offering_component_type': plan_component.component.type,
            'offering_component_name': plan_component.component.name,
            'offering_component_measured_unit': plan_component.component.measured_unit,
        })
        return details



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations

from waldur_mastermind.common.utils import quantize_price


class Migration(migrations.Migration):

    dependencies = [
        ('invoices', '0008_offeringitem'),
    ]

    def quantize_prices(apps, schema_editor):
        OpenStackItem = apps.get_model('invoices', 'OpenStackItem')
        for item in OpenStackItem.objects.iterator():
            item.daily_price = quantize_price(item.daily_price)
            item.save(update_fields=['daily_price'])

    operations = [
        migrations.RunPython(quantize_prices),
    ]



from waldur_core.logging.loggers import EventLogger, event_logger

from . import models


class IssueEventLogger(EventLogger):
    issue = models.Issue

    class Meta:
        event_types = ('issue_deletion_succeeded',
                       'issue_update_succeeded',
                       'issue_creation_succeeded')
        event_groups = {
            'support': event_types,
        }

    @staticmethod
    def get_scopes(event_context):
        issue = event_context['issue']
        result = set()
        if issue.resource:
            project = issue.resource.service_project_link.project
            result.add(issue.resource)
            result.add(project)
            result.add(project.customer)
        if issue.project:
            result.add(issue.project)
            result.add(issue.customer)
        if issue.customer:
            result.add(issue.customer)
        return result


class OfferingEventLogger(EventLogger):
    offering = models.Offering

    class Meta:
        event_types = (
            'offering_created',
            'offering_deleted',
            'offering_state_changed',
        )
        event_groups = {
            'support': event_types,
        }

    @staticmethod
    def get_scopes(event_context):
        offering = event_context['offering']
        return {offering.project, offering.project.customer}


event_logger.register('waldur_issue', IssueEventLogger)
event_logger.register('waldur_offering', OfferingEventLogger)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2018-10-02 07:24
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


def cleanup_all_offerings(apps, schema_editor):
    Offering = apps.get_model('marketplace', 'Offering')
    Offering.objects.all().delete()


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0037_component_usage'),
    ]

    operations = [
        migrations.RunPython(cleanup_all_offerings),
    ]



import six
from waldur_core.core.models import User
from waldur_core.logging.loggers import EventLogger, event_logger


class FreeIPAEventLogger(EventLogger):
    user = User
    username = six.text_type

    class Meta:
        event_types = (
            'freeipa_profile_created',
            'freeipa_profile_deleted',
            'freeipa_profile_enabled',
            'freeipa_profile_disabled',
        )
        event_groups = {'users': event_types}

    @staticmethod
    def get_scopes(event_context):
        return {event_context['user']}


event_logger.register('freeipa', FreeIPAEventLogger)



from __future__ import unicode_literals

from django.db import transaction
from django.utils.translation import ugettext_lazy as _
from rest_framework import serializers

from waldur_core.core import serializers as core_serializers
from waldur_core.structure import serializers as structure_serializers

from . import models
from .backend import RijkscloudBackendError


class ServiceSerializer(core_serializers.ExtraFieldOptionsMixin,
                        structure_serializers.BaseServiceSerializer):
    SERVICE_ACCOUNT_FIELDS = {
        'username': '',
        'token': '',
    }

    class Meta(structure_serializers.BaseServiceSerializer.Meta):
        model = models.RijkscloudService
        extra_field_options = {
            'username': {
                'label': 'User ID',
                'required': True
            },
            'token': {
                'label': 'API key',
                'required': True
            },
        }


class ServiceProjectLinkSerializer(structure_serializers.BaseServiceProjectLinkSerializer):
    class Meta(structure_serializers.BaseServiceProjectLinkSerializer.Meta):
        model = models.RijkscloudServiceProjectLink
        extra_kwargs = {
            'service': {'lookup_field': 'uuid', 'view_name': 'rijkscloud-detail'},
        }


class FlavorSerializer(structure_serializers.BasePropertySerializer):
    class Meta(structure_serializers.BasePropertySerializer.Meta):
        model = models.Flavor
        fields = ('url', 'uuid', 'name', 'settings', 'cores', 'ram',)
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
            'settings': {'lookup_field': 'uuid'},
        }


class VolumeSerializer(structure_serializers.BaseResourceSerializer):
    service = serializers.HyperlinkedRelatedField(
        source='service_project_link.service',
        view_name='rijkscloud-detail',
        read_only=True,
        lookup_field='uuid')

    service_project_link = serializers.HyperlinkedRelatedField(
        view_name='rijkscloud-spl-detail',
        queryset=models.RijkscloudServiceProjectLink.objects.all(),
        allow_null=True,
        required=False,
    )

    class Meta(structure_serializers.BaseResourceSerializer.Meta):
        model = models.Volume
        fields = structure_serializers.BaseResourceSerializer.Meta.fields + (
            'size', 'runtime_state')
        read_only_fields = structure_serializers.BaseResourceSerializer.Meta.read_only_fields + (
            'runtime_state',)
        protected_fields = structure_serializers.BaseResourceSerializer.Meta.protected_fields + (
            'size',)
        extra_kwargs = dict(
            size={'required': False, 'allow_null': True},
            **structure_serializers.BaseResourceSerializer.Meta.extra_kwargs
        )


class VolumeImportableSerializer(core_serializers.AugmentedSerializerMixin,
                                 serializers.HyperlinkedModelSerializer):
    service_project_link = serializers.HyperlinkedRelatedField(
        view_name='rijkscloud-spl-detail',
        queryset=models.RijkscloudServiceProjectLink.objects.all(),
        write_only=True)

    def get_filtered_field_names(self):
        return 'service_project_link',

    class Meta(object):
        model = models.Volume
        model_fields = ('name', 'description', 'size', 'runtime_state')
        fields = ('service_project_link', 'backend_id') + model_fields
        read_only_fields = model_fields


class VolumeImportSerializer(VolumeImportableSerializer):
    class Meta(VolumeImportableSerializer.Meta):
        fields = VolumeImportableSerializer.Meta.fields + ('url', 'uuid', 'created')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
        }

    @transaction.atomic
    def create(self, validated_data):
        service_project_link = validated_data['service_project_link']
        backend_id = validated_data['backend_id']

        if models.Volume.objects.filter(
            service_project_link__service__settings=service_project_link.service.settings,
            backend_id=backend_id
        ).exists():
            raise serializers.ValidationError({
                'backend_id': _('Volume has been imported already.')
            })

        try:
            backend = service_project_link.get_backend()
            volume = backend.import_volume(backend_id, save=True, service_project_link=service_project_link)
        except RijkscloudBackendError:
            raise serializers.ValidationError({
                'backend_id': _("Can't import volume with ID %s") % validated_data['backend_id']
            })

        return volume


class InstanceSerializer(structure_serializers.VirtualMachineSerializer):
    service = serializers.HyperlinkedRelatedField(
        source='service_project_link.service',
        view_name='rijkscloud-detail',
        read_only=True,
        lookup_field='uuid')

    service_project_link = serializers.HyperlinkedRelatedField(
        view_name='rijkscloud-spl-detail',
        queryset=models.RijkscloudServiceProjectLink.objects.all(),
        allow_null=True,
        required=False,
    )

    flavor = serializers.HyperlinkedRelatedField(
        view_name='rijkscloud-flavor-detail',
        lookup_field='uuid',
        queryset=models.Flavor.objects.all().select_related('settings'),
        write_only=True)

    floating_ip = serializers.HyperlinkedRelatedField(
        view_name='rijkscloud-fip-detail',
        lookup_field='uuid',
        queryset=models.FloatingIP.objects.filter(is_available=True).select_related('settings'),
        write_only=True,
        allow_null=True,
    )

    internal_ip = serializers.HyperlinkedRelatedField(
        view_name='rijkscloud-internal-ip-detail',
        lookup_field='uuid',
        queryset=models.InternalIP.objects.filter(is_available=True).select_related('settings'),
        write_only=True)

    class Meta(structure_serializers.VirtualMachineSerializer.Meta):
        model = models.Instance
        fields = structure_serializers.VirtualMachineSerializer.Meta.fields + (
            'flavor', 'floating_ip', 'internal_ip')
        protected_fields = structure_serializers.VirtualMachineSerializer.Meta.protected_fields + (
            'flavor', 'internal_ip',)
        read_only_fields = structure_serializers.VirtualMachineSerializer.Meta.read_only_fields + (
            'flavor_name',)

    def validate(self, attrs):
        attrs = super(InstanceSerializer, self).validate(attrs)

        # skip validation on object update
        if self.instance is not None:
            return attrs

        service_project_link = attrs['service_project_link']
        settings = service_project_link.service.settings
        flavor = attrs['flavor']

        if flavor.settings != settings:
            raise serializers.ValidationError(
                _('Flavor must belong to the same service settings as service project link.'))

        return attrs

    @transaction.atomic
    def create(self, validated_data):
        flavor = validated_data['flavor']
        validated_data['flavor_name'] = flavor.name
        validated_data['cores'] = flavor.cores
        validated_data['ram'] = flavor.ram

        floating_ip = validated_data.get('floating_ip')
        if floating_ip:
            floating_ip.is_available = False
            floating_ip.save()

        internal_ip = validated_data['internal_ip']
        internal_ip.is_available = False
        internal_ip.save()

        return super(InstanceSerializer, self).create(validated_data)


class InstanceImportableSerializer(core_serializers.AugmentedSerializerMixin,
                                   serializers.HyperlinkedModelSerializer):
    service_project_link = serializers.HyperlinkedRelatedField(
        view_name='rijkscloud-spl-detail',
        queryset=models.RijkscloudServiceProjectLink.objects.all(),
        write_only=True)

    def get_filtered_field_names(self):
        return 'service_project_link',

    class Meta(object):
        model = models.Instance
        model_fields = ('name', 'description', 'state', 'runtime_state', 'flavor_name', 'ram', 'cores')
        fields = ('service_project_link', 'backend_id') + model_fields
        read_only_fields = model_fields


class InstanceImportSerializer(InstanceImportableSerializer):
    class Meta(InstanceImportableSerializer.Meta):
        fields = InstanceImportableSerializer.Meta.fields + ('url', 'uuid', 'created')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
        }

    @transaction.atomic
    def create(self, validated_data):
        service_project_link = validated_data['service_project_link']
        backend_id = validated_data['backend_id']

        if models.Instance.objects.filter(
            service_project_link__service__settings=service_project_link.service.settings,
            backend_id=backend_id
        ).exists():
            raise serializers.ValidationError({'backend_id': _('Instance has been imported already.')})

        try:
            backend = service_project_link.get_backend()
            instance = backend.import_instance(backend_id, save=True, service_project_link=service_project_link)
        except RijkscloudBackendError:
            raise serializers.ValidationError({
                'backend_id': _("Can't import instance with ID %s") % validated_data['backend_id']
            })

        return instance


class NetworkSerializer(structure_serializers.BasePropertySerializer):
    class Meta(structure_serializers.BasePropertySerializer.Meta):
        model = models.Network
        fields = ('url', 'uuid', 'name', 'subnets')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
            'settings': {'lookup_field': 'uuid'},
            'subnets': {'lookup_field': 'uuid', 'view_name': 'rijkscloud-subnet-detail'}
        }


class SubNetSerializer(structure_serializers.BasePropertySerializer):
    dns_nameservers = serializers.JSONField(read_only=True)
    allocation_pools = serializers.JSONField(read_only=True)

    class Meta(structure_serializers.BasePropertySerializer.Meta):
        model = models.SubNet
        fields = ('url', 'uuid', 'name', 'cidr', 'gateway_ip', 'allocation_pools', 'dns_nameservers', 'network')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
            'settings': {'lookup_field': 'uuid'},
            'network': {'lookup_field': 'uuid', 'view_name': 'rijkscloud-network-detail'},
        }


class InternalIPSerializer(structure_serializers.BasePropertySerializer):
    class Meta(structure_serializers.BasePropertySerializer.Meta):
        model = models.InternalIP
        fields = ('url', 'uuid', 'subnet', 'address', 'is_available',)
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
            'subnet': {'lookup_field': 'uuid', 'view_name': 'rijkscloud-subnet-detail'},
        }


class FloatingIPSerializer(structure_serializers.BasePropertySerializer):
    class Meta(structure_serializers.BasePropertySerializer.Meta):
        model = models.FloatingIP
        fields = ('url', 'uuid', 'settings', 'address', 'is_available',)
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
            'settings': {'lookup_field': 'uuid'},
        }



from django.db import migrations
import uuid


def gen_uuid(apps, schema_editor):
    ComponentUsage = apps.get_model('marketplace', 'ComponentUsage')
    for row in ComponentUsage.objects.all():
        row.uuid = uuid.uuid4().hex
        row.save(update_fields=['uuid'])


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0069_componentusage_uuid'),
    ]

    operations = [
        migrations.RunPython(gen_uuid),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-03-26 17:56
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import django_fsm
import model_utils.fields
import waldur_core.core.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('structure', '0054_payment_details'),
        ('contenttypes', '0002_remove_content_type_name'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('python_management', '0003_added_unique_constraint'),
    ]

    operations = [
        migrations.CreateModel(
            name='JupyterHubManagement',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('session_time_to_live_hours', models.IntegerField(default=24)),
                ('instance_object_id', models.PositiveIntegerField(null=True)),
                ('instance_content_type', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='contenttypes.ContentType')),
            ],
        ),
        migrations.CreateModel(
            name='JupyterHubManagementDeleteRequest',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('state', django_fsm.FSMIntegerField(choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'), (4, 'Erred')], default=5)),
                ('output', models.TextField(blank=True)),
                ('jupyter_hub_management', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='jupyter_hub_management.JupyterHubManagement')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='JupyterHubManagementMakeVirtualEnvironmentGlobalRequest',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('state', django_fsm.FSMIntegerField(choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'), (4, 'Erred')], default=5)),
                ('virtual_env_name', models.CharField(max_length=255)),
                ('output', models.TextField(blank=True)),
                ('jupyter_hub_management', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='jupyter_hub_management.JupyterHubManagement')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='JupyterHubManagementMakeVirtualEnvironmentLocalRequest',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('state', django_fsm.FSMIntegerField(choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'), (4, 'Erred')], default=5)),
                ('virtual_env_name', models.CharField(max_length=255)),
                ('output', models.TextField(blank=True)),
                ('jupyter_hub_management', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='jupyter_hub_management.JupyterHubManagement')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='JupyterHubManagementSyncConfigurationRequest',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('state', django_fsm.FSMIntegerField(choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'), (4, 'Erred')], default=5)),
                ('output', models.TextField(blank=True)),
                ('jupyter_hub_management', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='jupyter_hub_management.JupyterHubManagement')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='JupyterHubOAuthConfig',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('type', models.PositiveSmallIntegerField(choices=[(1, 'GitLab'), (2, 'Microsoft Azure')])),
                ('oauth_callback_url', models.CharField(max_length=255)),
                ('client_id', models.CharField(max_length=255)),
                ('client_secret', models.CharField(max_length=255)),
                ('tenant_id', models.CharField(blank=True, max_length=255, null=True)),
                ('gitlab_host', models.CharField(blank=True, max_length=255, null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='JupyterHubUser',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('admin', models.BooleanField(default=False)),
                ('whitelisted', models.BooleanField(default=False)),
                ('username', models.CharField(max_length=255)),
                ('password', models.CharField(blank=True, max_length=255, null=True)),
                ('jupyter_hub_management', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='jupyter_hub_users', to='jupyter_hub_management.JupyterHubManagement')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='jupyterhubmanagementmakevirtualenvironmentglobalrequest',
            name='update_configuration_request',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='make_virtual_env_global_requests', to='jupyter_hub_management.JupyterHubManagementSyncConfigurationRequest'),
        ),
        migrations.AddField(
            model_name='jupyterhubmanagement',
            name='jupyter_hub_oauth_config',
            field=models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='jupyter_hub_management.JupyterHubOAuthConfig'),
        ),
        migrations.AddField(
            model_name='jupyterhubmanagement',
            name='project',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='structure.Project'),
        ),
        migrations.AddField(
            model_name='jupyterhubmanagement',
            name='python_management',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='python_management.PythonManagement'),
        ),
        migrations.AddField(
            model_name='jupyterhubmanagement',
            name='user',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterUniqueTogether(
            name='jupyterhubmanagement',
            unique_together=set([('instance_content_type', 'instance_object_id')]),
        ),
    ]



from __future__ import unicode_literals

from django.apps import AppConfig
from django.db.models import signals


class MarketplaceConfig(AppConfig):
    name = 'waldur_mastermind.marketplace'
    verbose_name = 'Marketplace'

    def ready(self):
        from waldur_core.quotas import signals as quota_signals

        from . import handlers, models

        signals.post_save.connect(
            handlers.create_screenshot_thumbnail,
            sender=models.Screenshot,
            dispatch_uid='waldur_mastermind.marketplace.create_screenshot_thumbnail',
        )

        signals.post_save.connect(
            handlers.log_order_events,
            sender=models.Order,
            dispatch_uid='waldur_mastermind.marketplace.log_order_events',
        )

        signals.post_save.connect(
            handlers.log_order_item_events,
            sender=models.OrderItem,
            dispatch_uid='waldur_mastermind.marketplace.log_order_item_events',
        )

        signals.post_save.connect(
            handlers.log_resource_events,
            sender=models.Resource,
            dispatch_uid='waldur_mastermind.marketplace.log_resource_events',
        )

        signals.post_save.connect(
            handlers.reject_order,
            sender=models.Order,
            dispatch_uid='waldur_mastermind.marketplace.reject_order',
        )

        signals.post_save.connect(
            handlers.complete_order_when_all_items_are_done,
            sender=models.OrderItem,
            dispatch_uid='waldur_mastermind.marketplace.complete_order_when_all_items_are_done',
        )

        signals.post_save.connect(
            handlers.update_category_quota_when_offering_is_created,
            sender=models.Offering,
            dispatch_uid='waldur_mastermind.marketplace.update_category_quota_when_offering_is_created',
        )

        signals.post_delete.connect(
            handlers.update_category_quota_when_offering_is_deleted,
            sender=models.Offering,
            dispatch_uid='waldur_mastermind.marketplace.update_category_quota_when_offering_is_deleted',
        )

        quota_signals.recalculate_quotas.connect(
            handlers.update_category_offerings_count,
            dispatch_uid='waldur_mastermind.marketplace.update_category_offerings_count',
        )

        signals.post_save.connect(
            handlers.update_aggregate_resources_count_when_resource_is_updated,
            sender=models.Resource,
            dispatch_uid='waldur_mastermind.marketplace.'
                         'update_aggregate_resources_count_when_resource_is_updated',
        )

        quota_signals.recalculate_quotas.connect(
            handlers.update_aggregate_resources_count,
            dispatch_uid='waldur_mastermind.marketplace.update_aggregate_resources_count',
        )

        signals.post_save.connect(
            handlers.close_resource_plan_period_when_resource_is_terminated,
            sender=models.Resource,
            dispatch_uid='waldur_mastermind.marketplace.'
                         'close_resource_plan_period_when_resource_is_terminated',
        )



import django_filters

from waldur_core.core import filters as core_filters
from waldur_core.structure import models as structure_models, filters as structure_filters

from . import models


class IssueFilter(django_filters.FilterSet):
    summary = django_filters.CharFilter(lookup_expr='icontains')

    customer = core_filters.URLFilter(view_name='customer-detail', name='customer__uuid')
    customer_uuid = django_filters.UUIDFilter(name='customer__uuid')

    project = core_filters.URLFilter(view_name='project-detail', name='project__uuid')
    project_uuid = django_filters.UUIDFilter(name='project__uuid')

    reporter_name = django_filters.CharFilter(lookup_expr='icontains', name='reporter__name')
    reporter = core_filters.URLFilter(view_name='support-user-detail', name='reporter__uuid')

    caller_full_name = django_filters.CharFilter(lookup_expr='icontains', name='caller__full_name')
    caller = core_filters.URLFilter(view_name='user-detail', name='caller__uuid')

    assignee_name = django_filters.CharFilter(lookup_expr='icontains', name='assignee__name')
    assignee = core_filters.URLFilter(view_name='support-user-detail', name='assignee__uuid')

    o = django_filters.OrderingFilter(
        fields=(
            ('created', 'created'),
            ('modified', 'modified'),
            ('type', 'type'),
            ('key', 'key'),
            ('status', 'status'),
            ('priority', 'priority'),
            ('summary', 'summary'),
            ('customer__name', 'customer_name'),
            ('project__name', 'project_name'),
            ('caller__full_name', 'caller_full_name'),
            ('reporter__name', 'reporter_name'),
            ('assignee__name', 'assignee_name'),
        ))

    class Meta(object):
        model = models.Issue
        fields = [
            'key',
            'type',
            'status',
        ]


class IssueResourceFilterBackend(core_filters.GenericKeyFilterBackend):

    content_type_field = 'resource_content_type'
    object_id_field = 'resource_object_id'

    def get_related_models(self):
        return structure_models.ResourceMixin.get_all_models()

    def get_field_name(self):
        return 'resource'


class CommentIssueResourceFilterBackend(IssueResourceFilterBackend):

    content_type_field = 'issue__resource_content_type'
    object_id_field = 'issue__resource_object_id'


class IssueCallerOrRoleFilterBackend(structure_filters.GenericRoleFilter):
    def filter_queryset(self, request, queryset, view):
        return super(IssueCallerOrRoleFilterBackend, self).filter_queryset(request, queryset, view).distinct() | \
            queryset.filter(caller=request.user).distinct()


class CommentIssueCallerOrRoleFilterBackend(structure_filters.GenericRoleFilter):
    def filter_queryset(self, request, queryset, view):
        return super(CommentIssueCallerOrRoleFilterBackend, self).filter_queryset(request,
                                                                                  queryset,
                                                                                  view).distinct() | \
            queryset.filter(issue__caller=request.user).distinct()


class CommentFilter(django_filters.FilterSet):
    description = django_filters.CharFilter(lookup_expr='icontains')
    issue = core_filters.URLFilter(view_name='support-issue-detail', name='issue__uuid')
    issue_uuid = django_filters.UUIDFilter(name='issue__uuid')
    author_name = django_filters.CharFilter(lookup_expr='icontains', name='author__name')
    author_user = core_filters.URLFilter(view_name='user-detail', name='author__user__uuid')

    o = django_filters.OrderingFilter(fields=('created', 'modified'))

    class Meta(object):
        model = models.Comment
        fields = [
            'is_public',
        ]


class SupportUserFilter(django_filters.FilterSet):
    name = django_filters.CharFilter(lookup_expr='icontains')

    class Meta(object):
        model = models.SupportUser
        fields = ('name', 'user', 'backend_id')


class OfferingFilter(django_filters.FilterSet):
    name = django_filters.CharFilter(lookup_expr='icontains')
    description = django_filters.CharFilter(lookup_expr='icontains')
    type = django_filters.ModelMultipleChoiceFilter(queryset=models.OfferingTemplate.objects.all(),
                                                    to_field_name='name',
                                                    method='offering_template_filter')
    template = django_filters.ModelMultipleChoiceFilter(queryset=models.OfferingTemplate.objects.all(), )
    issue = core_filters.URLFilter(view_name='support-issue-detail', name='issue__uuid')
    issue_uuid = django_filters.UUIDFilter(name='issue__uuid')
    issue_key = django_filters.CharFilter(name='issue__key')
    project = core_filters.URLFilter(view_name='project-detail', name='project__uuid')
    project_uuid = django_filters.UUIDFilter(name='project__uuid')
    state = core_filters.MappedMultipleChoiceFilter(
        choices=[(representation, representation) for db_value, representation in models.Offering.States.CHOICES],
        choice_mappings={representation: db_value for db_value, representation in models.Offering.States.CHOICES},
    )

    o = django_filters.OrderingFilter(fields=('created', 'modified', 'state'))

    def offering_template_filter(self, queryset, name, value):
        if value:
            return queryset.filter(template__in=value)
        return queryset

    class Meta(object):
        model = models.Offering
        fields = ('name', 'description', 'template', 'issue', 'issue_uuid', 'project', 'project_uuid', 'state')


class AttachmentFilter(django_filters.FilterSet):
    issue = core_filters.URLFilter(view_name='support-issue-detail', name='issue__uuid')
    issue_uuid = django_filters.UUIDFilter(name='issue__uuid')

    class Meta(object):
        model = models.Attachment
        fields = ('issue', 'issue_uuid')



from django.apps import AppConfig
from django.db.models import signals


class PackageConfig(AppConfig):
    name = 'waldur_mastermind.packages'
    verbose_name = 'VPC packages'

    def ready(self):
        OpenStackPackage = self.get_model('OpenStackPackage')

        from waldur_mastermind.invoices import registrators
        from . import handlers, registrators as openstack_registrator
        from . import cost_planning  # noqa: F401

        signals.post_save.connect(
            handlers.log_openstack_package_creation,
            sender=OpenStackPackage,
            dispatch_uid='waldur_mastermind.packages.log_openstack_package_creation',
        )

        signals.pre_delete.connect(
            handlers.log_openstack_package_deletion,
            sender=OpenStackPackage,
            dispatch_uid='waldur_mastermind.packages.log_openstack_package_deletion',
        )

        registrators.RegistrationManager.add_registrator(
            OpenStackPackage,
            openstack_registrator.OpenStackItemRegistrator
        )

        signals.post_save.connect(
            handlers.add_new_openstack_package_details_to_invoice,
            sender=OpenStackPackage,
            dispatch_uid='waldur_mastermind.invoices.add_new_openstack_package_details_to_invoice',
        )

        signals.pre_delete.connect(
            handlers.update_invoice_on_openstack_package_deletion,
            sender=OpenStackPackage,
            dispatch_uid='waldur_mastermind.invoices.update_invoice_on_openstack_package_deletion',
        )



from waldur_core.core import executors as core_executors, tasks as core_tasks


class PythonManagementRequestExecutor(core_executors.CreateExecutor):
    @classmethod
    def get_task_signature(cls, python_management, serialized_python_management_request, **kwargs):
        return core_tasks.BackendMethodTask().si(
            serialized_python_management_request, 'process_python_management_request', state_transition='begin_creating')



# -*- coding: utf-8 -*-
# Generated by Django 1.11.18 on 2019-03-26 08:14
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0071_componentusage_uuid_unique'),
    ]

    operations = [
        migrations.AddField(
            model_name='offeringcomponent',
            name='article_code',
            field=models.CharField(blank=True, max_length=30),
        ),
        migrations.AddField(
            model_name='offeringcomponent',
            name='product_code',
            field=models.CharField(blank=True, max_length=30),
        ),
    ]



import logging

from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist, MultipleObjectsReturned
from django.db import transaction

from waldur_core.core import models as core_models
from waldur_core.core.utils import serialize_instance
from waldur_core.structure import models as structure_models
from waldur_mastermind.marketplace import models as marketplace_models, plugins
from waldur_mastermind.marketplace.utils import import_resource_metadata, format_list
from waldur_mastermind.marketplace_openstack import (
    INSTANCE_TYPE, VOLUME_TYPE, PACKAGE_TYPE, RAM_TYPE, STORAGE_TYPE
)
from waldur_mastermind.packages import models as package_models
from waldur_openstack.openstack import apps as openstack_apps
from waldur_openstack.openstack import models as openstack_models
from waldur_openstack.openstack_tenant import apps as openstack_tenant_apps
from waldur_openstack.openstack_tenant import models as openstack_tenant_models

logger = logging.getLogger(__name__)


def get_offering_category_for_tenant():
    return marketplace_models.Category.objects.get(
        uuid=settings.WALDUR_MARKETPLACE_OPENSTACK['TENANT_CATEGORY_UUID']
    )


def get_offering_name_for_instance(tenant):
    return 'Virtual machine in %s' % tenant.name


def get_offering_category_for_instance():
    return marketplace_models.Category.objects.get(
        uuid=settings.WALDUR_MARKETPLACE_OPENSTACK['INSTANCE_CATEGORY_UUID']
    )


def get_offering_name_for_volume(tenant):
    return 'Volume in %s' % tenant.name


def get_offering_category_for_volume():
    return marketplace_models.Category.objects.get(
        uuid=settings.WALDUR_MARKETPLACE_OPENSTACK['VOLUME_CATEGORY_UUID']
    )


def get_category_and_name_for_offering_type(offering_type, service_settings):
    if offering_type == INSTANCE_TYPE:
        category = get_offering_category_for_instance()
        name = get_offering_name_for_instance(service_settings)
        return category, name
    elif offering_type == VOLUME_TYPE:
        category = get_offering_category_for_volume()
        name = get_offering_name_for_volume(service_settings)
        return category, name


def get_resource_state(state):
    SrcStates = core_models.StateMixin.States
    DstStates = marketplace_models.Resource.States
    mapping = {
        SrcStates.CREATION_SCHEDULED: DstStates.CREATING,
        SrcStates.CREATING: DstStates.CREATING,
        SrcStates.UPDATE_SCHEDULED: DstStates.UPDATING,
        SrcStates.UPDATING: DstStates.UPDATING,
        SrcStates.DELETION_SCHEDULED: DstStates.TERMINATING,
        SrcStates.DELETING: DstStates.TERMINATING,
        SrcStates.OK: DstStates.OK,
        SrcStates.ERRED: DstStates.ERRED,
    }
    return mapping.get(state, DstStates.ERRED)


def create_offering_components(offering):
    fixed_components = plugins.manager.get_components(PACKAGE_TYPE)

    for component_data in fixed_components:
        marketplace_models.OfferingComponent.objects.create(
            offering=offering,
            **component_data._asdict()
        )


def copy_plan_components_from_template(plan, offering, template):
    component_map = {
        component.type: component
        for component in template.components.all()
    }

    for (key, component_data) in component_map.items():
        plan_component = component_map.get(key)
        offering_component = offering.components.get(type=key)

        amount = plan_component.amount
        price = plan_component.price

        # In marketplace RAM and storage is stored in GB, but in package plugin it is stored in MB.
        if key in (RAM_TYPE, STORAGE_TYPE):
            amount = int(amount / 1024)
            price = price * 1024

        marketplace_models.PlanComponent.objects.create(
            plan=plan,
            component=offering_component,
            amount=amount,
            price=price,
        )


def import_openstack_service_settings(default_customer, dry_run=False, require_templates=False):
    """
    Import OpenStack service settings as marketplace offerings.
    """
    service_type = openstack_apps.OpenStackConfig.service_name
    category = get_offering_category_for_tenant()

    package_offerings = marketplace_models.Offering.objects.filter(type=PACKAGE_TYPE)
    front_settings = set(package_offerings.exclude(object_id=None).values_list('object_id', flat=True))

    back_settings = structure_models.ServiceSettings.objects.filter(type=service_type)
    missing_settings = back_settings.exclude(id__in=front_settings)

    if dry_run:
        logger.warning('OpenStack service settings would be imported to marketplace. '
                       'ID: %s.', format_list(missing_settings))
        return 0, 0

    missing_templates = package_models.PackageTemplate.objects.filter(
        service_settings__in=missing_settings)

    settings_without_templates = missing_settings.exclude(
        id__in=missing_templates.values_list('service_settings_id', flat=True))

    def create_offering(service_settings, state):
        offering = marketplace_models.Offering.objects.create(
            scope=service_settings,
            type=PACKAGE_TYPE,
            name=service_settings.name,
            geolocations=service_settings.geolocations,
            customer=service_settings.customer or default_customer,
            category=category,
            shared=service_settings.shared,
            state=state,
        )
        create_offering_components(offering)
        return offering

    offerings_counter = 0
    plans_counter = 0

    if settings_without_templates.exists():
        logger.warning('The following service settings do not have package template, '
                       'therefore they would be imported in DRAFT state: %s',
                       format_list(settings_without_templates))

    if not require_templates:
        for service_settings in settings_without_templates:
            with transaction.atomic():
                create_offering(service_settings, marketplace_models.Offering.States.DRAFT)
                offerings_counter += 1

    for template in missing_templates:
        with transaction.atomic():
            service_settings = template.service_settings

            try:
                offering = marketplace_models.Offering.objects.get(scope=service_settings)
            except marketplace_models.Offering.DoesNotExist:
                offering = create_offering(service_settings, marketplace_models.Offering.States.ACTIVE)
                offerings_counter += 1

            plan = marketplace_models.Plan.objects.create(
                offering=offering,
                name=template.name,
                unit_price=template.price,
                unit=marketplace_models.Plan.Units.PER_DAY,
                product_code=template.product_code,
                article_code=template.article_code,
                scope=template,
            )
            plans_counter += 1

            copy_plan_components_from_template(plan, offering, template)

    return offerings_counter, plans_counter


def import_openstack_tenants(dry_run=False):
    """
    Import OpenStack tenants as marketplace resources.
    It is expected that offerings for OpenStack service settings are imported before this command is ran.
    """
    front_ids = set(marketplace_models.Resource.objects.
                    filter(offering__type=PACKAGE_TYPE).
                    values_list('object_id', flat=True))
    missing_resources = openstack_models.Tenant.objects.exclude(id__in=front_ids)

    if dry_run:
        logger.warning('OpenStack tenants would be imported to marketplace. '
                       'ID: %s.', format_list(missing_resources))
        return 0

    packages = package_models.OpenStackPackage.objects.filter(tenant__in=missing_resources)
    tenants_without_packages = missing_resources.exclude(id__in=packages.values_list('tenant_id', flat=True))

    def create_resource(offering, tenant, plan=None):
        resource = marketplace_models.Resource.objects.create(
            name=tenant.name,
            created=tenant.created,
            offering=offering,
            plan=plan,
            scope=tenant,
            project=tenant.project,
            state=get_resource_state(tenant.state),
            attributes=dict(
                name=tenant.name,
                description=tenant.description,
                user_username=tenant.user_username,
                user_password=tenant.user_password,
            )
        )
        if plan and tenant.backend_id:
            marketplace_models.ResourcePlanPeriod.objects.create(
                resource=resource,
                plan=plan,
                start=tenant.created,
            )
        import_resource_metadata(resource)
        return resource

    resource_counter = 0
    for tenant in tenants_without_packages:
        # It is expected that service setting has exactly one offering
        # if it does not have package
        try:
            offering = marketplace_models.Offering.objects.get(scope=tenant.service_settings)
        except marketplace_models.Offering.DoesNotExist:
            logger.warning('Offering for service setting is not imported yet. '
                           'Service setting ID: %s.', tenant.service_settings.id)
            continue

        create_resource(offering, tenant)
        resource_counter += 1

    for package in packages:
        tenant = package.tenant
        try:
            offering = marketplace_models.Offering.objects.get(scope=tenant.service_settings)
            plan = marketplace_models.Plan.objects.get(scope=package.template, offering=offering)
        except marketplace_models.Plan.DoesNotExist:
            logger.warning('Plan for template is not imported yet. '
                           'Template ID: %s.', package.template_id)
            continue

        create_resource(plan.offering, tenant, plan)
        resource_counter += 1

    return resource_counter


def import_openstack_tenant_service_settings(dry_run=False):
    """
    Import OpenStack tenant service settings as marketplace offerings.
    """

    offerings_counter = 0
    plans_counter = 0

    for offering_type in (INSTANCE_TYPE, VOLUME_TYPE):
        marketplace_offerings = marketplace_models.Offering.objects.filter(type=offering_type)
        front_settings = set(marketplace_offerings.exclude(object_id=None).values_list('object_id', flat=True))
        missing_settings = structure_models.ServiceSettings.objects.filter(
            type=openstack_tenant_apps.OpenStackTenantConfig.service_name
        ).exclude(id__in=front_settings)

        if dry_run:
            logger.warning('OpenStack tenant service settings would be imported to marketplace. '
                           'ID: %s.', format_list(missing_settings))
            continue

        packages = package_models.OpenStackPackage.objects.filter(service_settings__in=missing_settings)
        settings_to_template = {package.service_settings: package.template for package in packages}

        for service_settings in missing_settings:
            category, offering_name = get_category_and_name_for_offering_type(offering_type, service_settings)
            offering = marketplace_models.Offering.objects.create(
                customer=service_settings.customer,
                category=category,
                name=offering_name,
                scope=service_settings,
                shared=service_settings.shared,
                type=offering_type,
                state=marketplace_models.Offering.States.ACTIVE,
                billable=False,
            )
            create_offering_components(offering)
            offerings_counter += 1

            template = settings_to_template.get(service_settings)
            if not template:
                logger.warning('Billing for service setting is not imported because it does not have template. '
                               'Service setting ID: %s', service_settings.id)
                continue

            try:
                parent_plan = marketplace_models.Plan.objects.get(scope=template, offering__type=PACKAGE_TYPE)
            except marketplace_models.Plan.DoesNotExist:
                logger.warning('Billing for template is not imported because it does not have plan. '
                               'Template ID: %s', template.id)
                continue

            plan = marketplace_models.Plan.objects.create(
                offering=offering,
                name=parent_plan.name,
                scope=parent_plan.scope
            )

            copy_plan_components_from_template(plan, offering, template)
            plans_counter += 1

    return offerings_counter, plans_counter


def get_plan_for_resource(resource, offering):
    tenant = resource.service_settings.scope
    if not tenant:
        logger.warning('Skipping billing for resource because it does not have shared OpenStack settings. '
                       'Resource: %s', serialize_instance(resource))
        return

    try:
        package = package_models.OpenStackPackage.objects.get(tenant=tenant)
    except package_models.OpenStackPackage.DoesNotExist:
        logger.warning('Skipping billing for resource because package for tenant is not defined. '
                       'Tenant ID: %s', tenant.id)
        return

    try:
        plan = marketplace_models.Plan.objects.get(scope=package.template, offering=offering)
    except marketplace_models.Plan.DoesNotExist:
        logger.warning('Skipping billing for resource because plan for template is not defined. '
                       'Template ID: %s', package.template)
        return

    return plan


def import_openstack_instances_and_volumes(dry_run=False):
    """
    Import OpenStack tenant resources as marketplace resources.
    It is expected that offerings for OpenStack tenant service settings are imported before this command is ran.
    """
    model_classes = {
        INSTANCE_TYPE: openstack_tenant_models.Instance,
        VOLUME_TYPE: openstack_tenant_models.Volume,
    }

    resources_counter = 0

    for offering_type in (INSTANCE_TYPE, VOLUME_TYPE):
        front_ids = set(marketplace_models.Resource.objects.
                        filter(offering__type=offering_type).
                        values_list('object_id', flat=True))

        model_class = model_classes[offering_type]
        missing_resources = model_class.objects.exclude(id__in=front_ids)

        if dry_run:
            ids = format_list(missing_resources)
            logger.warning('OpenStack resource with IDs would be imported to marketplace: %s.', ids)
            continue

        offerings = {
            offering.scope: offering
            for offering in marketplace_models.Offering.objects.filter(type=offering_type)
        }

        for resource in missing_resources:
            offering = offerings.get(resource.service_settings)
            if not offering:
                logger.warning('Offering for service setting with ID %s is not imported yet.',
                               resource.service_settings.id)
                continue

            plan = get_plan_for_resource(resource, offering)

            new_resource = marketplace_models.Resource.objects.create(
                name=resource.name,
                created=resource.created,
                project=resource.project,
                offering=offering,
                plan=plan,
                scope=resource,
                state=get_resource_state(resource.state),
                attributes=dict(
                    name=resource.name,
                    description=resource.description,
                ),
            )
            if isinstance(resource, openstack_tenant_models.Volume):
                import_volume_metadata(new_resource)
            if isinstance(resource, openstack_tenant_models.Instance):
                import_instance_metadata(new_resource)
            resources_counter += 1

    return resources_counter


def import_volume_metadata(resource):
    import_resource_metadata(resource)
    volume = resource.scope
    resource.backend_metadata['size'] = volume.size

    if volume.instance:
        resource.backend_metadata['instance_uuid'] = volume.instance.uuid.hex
        resource.backend_metadata['instance_name'] = volume.instance.name
    else:
        resource.backend_metadata['instance_uuid'] = None
        resource.backend_metadata['instance_name'] = None

    resource.save(update_fields=['backend_metadata'])


def import_instance_metadata(resource):
    import_resource_metadata(resource)
    instance = resource.scope
    resource.backend_metadata['internal_ips'] = instance.internal_ips
    resource.backend_metadata['external_ips'] = instance.external_ips
    resource.save(update_fields=['backend_metadata'])


def get_offering(offering_type, service_settings):
    try:
        return marketplace_models.Offering.objects.get(scope=service_settings, type=offering_type)
    except ObjectDoesNotExist:
        logger.warning('Marketplace offering is not found. '
                       'ServiceSettings ID: %s', service_settings.id)
    except MultipleObjectsReturned:
        logger.warning('Few marketplace offerings are found. '
                       'ServiceSettings ID: %s', service_settings.id)



import math

from waldur_mastermind.invoices import registrators
from waldur_slurm.structures import Quotas


def get_price(quotas, package):
    minutes_in_hour = 60
    mb_in_gb = 1024
    cpu_price = int(math.ceil(1.0 * quotas.cpu / minutes_in_hour)) * package.cpu_price
    gpu_price = int(math.ceil(1.0 * quotas.gpu / minutes_in_hour)) * package.gpu_price
    ram_price = int(math.ceil(1.0 * quotas.ram / mb_in_gb)) * package.ram_price
    return cpu_price + gpu_price + ram_price


def get_package(allocation):
    registrator = registrators.RegistrationManager.get_registrator(allocation)
    return registrator.get_package(allocation)


def get_deposit_limit(allocation, package):
    quotas = Quotas(allocation.cpu_limit, allocation.gpu_limit, allocation.ram_limit)
    return get_price(quotas, package)


def get_deposit_usage(allocation, package):
    if allocation.batch_service == 'MOAB':
        return allocation.deposit_usage
    else:
        quotas = Quotas(allocation.cpu_usage, allocation.gpu_usage, allocation.ram_usage)
        return get_price(quotas, package)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-06-19 08:47
from __future__ import unicode_literals

from django.db import migrations, models


def migrate_data(apps, schema_editor):
    Invoice = apps.get_model('invoices', 'Invoice')

    for invoice in Invoice.objects.all():
        invoice.update_current_cost()


class Migration(migrations.Migration):

    dependencies = [
        ('invoices', '0022_remove_payment_details'),
    ]

    operations = [
        migrations.AddField(
            model_name='invoice',
            name='current_cost',
            field=models.DecimalField(decimal_places=2, default=0, editable=False, help_text='Cached value for current cost.', max_digits=10),
        ),
        migrations.RunPython(migrate_data, reverse_code=migrations.RunPython.noop),
    ]



from django.core.management.base import BaseCommand
from django.db import transaction
from django.utils import timezone

from waldur_core.core import utils as core_utils
from waldur_core.cost_tracking import models, CostTrackingRegister, tasks


class Command(BaseCommand):
    help = ("Delete all price estimates that are related to current month and "
            "create new ones based on current consumption.")

    def handle(self, *args, **options):
        today = timezone.now()
        with transaction.atomic():
            # Delete current month price estimates
            models.PriceEstimate.objects.filter(month=today.month, year=today.year).delete()
            # Create new estimates for resources and ancestors
            for resource_model in CostTrackingRegister.registered_resources:
                for resource in resource_model.objects.all():
                    configuration = CostTrackingRegister.get_configuration(resource)
                    date = max(core_utils.month_start(today), resource.created)
                    models.PriceEstimate.create_historical(resource, configuration, date)
            # recalculate consumed estimate
            tasks.recalculate_estimate()



from django import forms
from django.contrib import admin
from django.core.exceptions import ValidationError
from django.utils.translation import ugettext_lazy as _
import pytz

from waldur_core.core.admin import ExecutorAdminAction, format_json_field
from waldur_core.structure import admin as structure_admin

from . import executors, models


class FlavorAdmin(structure_admin.BackendModelAdmin):
    list_filter = ('settings',)
    list_display = ('name', 'settings', 'cores', 'ram', 'disk')


class ImageAdmin(structure_admin.BackendModelAdmin):
    list_filter = ('settings',)
    list_display = ('name', 'settings', 'min_disk', 'min_ram')


class FloatingIPAdmin(structure_admin.BackendModelAdmin):
    list_filter = ('settings',)
    list_display = ('address', 'settings', 'runtime_state', 'backend_network_id', 'is_booked')


class SecurityGroupRule(admin.TabularInline):
    model = models.SecurityGroupRule
    fields = ('protocol', 'from_port', 'to_port', 'cidr', 'backend_id')
    readonly_fields = fields
    extra = 0
    can_delete = False


class SecurityGroupAdmin(structure_admin.BackendModelAdmin):
    inlines = [SecurityGroupRule]
    list_filter = ('settings',)
    list_display = ('name', 'settings')


class MetadataMixin(admin.ModelAdmin):
    def get_readonly_fields(self, request, obj=None):
        return super(MetadataMixin, self).get_readonly_fields(request, obj) + ('format_metadata',)

    def format_metadata(self, obj):
        return format_json_field(obj.metadata)

    format_metadata.allow_tags = True
    format_metadata.short_description = _('Metadata')


class ImageMetadataMixin(admin.ModelAdmin):
    def get_readonly_fields(self, request, obj=None):
        return super(ImageMetadataMixin, self).get_readonly_fields(request, obj) + ('format_image_metadata',)

    def format_image_metadata(self, obj):
        return format_json_field(obj.image_metadata)

    format_image_metadata.allow_tags = True
    format_image_metadata.short_description = _('Image metadata')


class ActionDetailsMixin(admin.ModelAdmin):
    def get_readonly_fields(self, request, obj=None):
        return super(ActionDetailsMixin, self).get_readonly_fields(request, obj) + ('format_action_details',)

    def format_action_details(self, obj):
        return format_json_field(obj.action_details)

    format_action_details.allow_tags = True
    format_action_details.short_description = _('Action details')


class VolumeAdmin(MetadataMixin,
                  ImageMetadataMixin,
                  ActionDetailsMixin,
                  structure_admin.ResourceAdmin):
    exclude = ('metadata', 'image_metadata', 'action_details')

    class Pull(ExecutorAdminAction):
        executor = executors.VolumePullExecutor
        short_description = _('Pull')

        def validate(self, instance):
            if instance.state not in (models.Volume.States.OK, models.Volume.States.ERRED):
                raise ValidationError(_('Volume has to be in OK or ERRED state.'))

    pull = Pull()


class SnapshotAdmin(structure_admin.ResourceAdmin):
    class Pull(ExecutorAdminAction):
        executor = executors.SnapshotPullExecutor
        short_description = _('Pull')

        def validate(self, instance):
            if instance.state not in (models.Snapshot.States.OK, models.Snapshot.States.ERRED):
                raise ValidationError(_('Snapshot has to be in OK or ERRED state.'))

    pull = Pull()


class InternalIpInline(admin.TabularInline):
    model = models.InternalIP

    def has_add_permission(self, request):
        return False

    def get_readonly_fields(self, request, obj=None):
        return models.InternalIP.get_backend_fields() + ('backend_id', 'instance', 'subnet')


class InstanceAdmin(ActionDetailsMixin, structure_admin.VirtualMachineAdmin):
    actions = structure_admin.VirtualMachineAdmin.actions + ['pull']
    exclude = ('action_details',)
    inlines = [InternalIpInline]
    list_filter = structure_admin.VirtualMachineAdmin.list_filter + ('runtime_state',)
    search_fields = structure_admin.VirtualMachineAdmin.search_fields + ('uuid', 'backend_id', 'runtime_state')

    class Pull(ExecutorAdminAction):
        executor = executors.InstancePullExecutor
        short_description = _('Pull')

        def validate(self, instance):
            if instance.state not in (models.Instance.States.OK, models.Instance.States.ERRED):
                raise ValidationError(_('Instance has to be in OK or ERRED state.'))

    pull = Pull()


class BackupAdmin(MetadataMixin, admin.ModelAdmin):
    readonly_fields = ('created', 'kept_until')
    list_filter = ('uuid', 'state')
    list_display = ('uuid', 'instance', 'state', 'project')
    exclude = ('metadata',)

    def project(self, obj):
        return obj.instance.service_project_link.project

    project.short_description = _('Project')


class BaseScheduleForm(forms.ModelForm):
    def clean_timezone(self):
        tz = self.cleaned_data['timezone']
        if tz not in pytz.all_timezones:
            raise ValidationError(_('Invalid timezone'), code='invalid')

        return self.cleaned_data['timezone']


class BaseScheduleAdmin(structure_admin.ResourceAdmin):
    form = BaseScheduleForm
    readonly_fields = ('next_trigger_at',)
    list_filter = ('is_active',) + structure_admin.ResourceAdmin.list_filter
    list_display = ('uuid', 'next_trigger_at', 'is_active', 'timezone') + structure_admin.ResourceAdmin.list_display


class BackupScheduleAdmin(BaseScheduleAdmin):
    list_display = BaseScheduleAdmin.list_display + ('instance',)


class SnapshotScheduleAdmin(BaseScheduleAdmin):
    list_display = BaseScheduleAdmin.list_display + ('source_volume',)


admin.site.register(models.OpenStackTenantService, structure_admin.ServiceAdmin)
admin.site.register(models.OpenStackTenantServiceProjectLink, structure_admin.ServiceProjectLinkAdmin)
admin.site.register(models.Flavor, FlavorAdmin)
admin.site.register(models.Image, ImageAdmin)
admin.site.register(models.FloatingIP, FloatingIPAdmin)
admin.site.register(models.SecurityGroup, SecurityGroupAdmin)
admin.site.register(models.Volume, VolumeAdmin)
admin.site.register(models.VolumeType, structure_admin.ServicePropertyAdmin)
admin.site.register(models.VolumeAvailabilityZone, structure_admin.ServicePropertyAdmin)
admin.site.register(models.Snapshot, SnapshotAdmin)
admin.site.register(models.InstanceAvailabilityZone, structure_admin.ServicePropertyAdmin)
admin.site.register(models.Instance, InstanceAdmin)
admin.site.register(models.Backup, BackupAdmin)
admin.site.register(models.BackupSchedule, BackupScheduleAdmin)
admin.site.register(models.SnapshotSchedule, SnapshotScheduleAdmin)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-07-03 17:02
from __future__ import unicode_literals

from django.db import migrations
import waldur_core.core.fields


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0001_squashed_0054'),
    ]

    operations = [
        migrations.AlterField(
            model_name='servicesettings',
            name='geolocations',
            field=waldur_core.core.fields.JSONField(blank=True, default=list, help_text='List of latitudes and longitudes. For example: [{"latitude": 123, "longitude": 345}, {"latitude": 456, "longitude": 678}]'),
        ),
        migrations.AlterField(
            model_name='servicesettings',
            name='options',
            field=waldur_core.core.fields.JSONField(blank=True, default=dict, help_text='Extra options'),
        ),
    ]



default_app_config = 'waldur_mastermind.marketplace_openstack.apps.MarketplaceOpenStackConfig'

PACKAGE_TYPE = 'Packages.Template'
INSTANCE_TYPE = 'OpenStackTenant.Instance'
VOLUME_TYPE = 'OpenStackTenant.Volume'

RAM_TYPE = 'ram'
CORES_TYPE = 'cores'
STORAGE_TYPE = 'storage'



from __future__ import unicode_literals

import logging
from smtplib import SMTPException

from celery import shared_task
from django.conf import settings
from django.utils import timezone
from python_freeipa import exceptions as freeipa_exceptions
import requests

from waldur_core.core import models as core_models
from waldur_core.core.utils import broadcast_mail, pwgen
from waldur_core.users import models, utils
from waldur_core.users.utils import generate_safe_username

logger = logging.getLogger(__name__)


@shared_task(name='waldur_core.users.cancel_expired_invitations')
def cancel_expired_invitations(invitations=None):
    """
    Invitation lifetime must be specified in Waldur Core settings with parameter
    "INVITATION_LIFETIME". If invitation creation time is less than expiration time, the invitation will set as expired.
    """
    expiration_date = timezone.now() - settings.WALDUR_CORE['INVITATION_LIFETIME']
    if not invitations:
        invitations = models.Invitation.objects.filter(state=models.Invitation.State.PENDING)
    invitations = invitations.filter(created__lte=expiration_date)
    invitations.update(state=models.Invitation.State.EXPIRED)


@shared_task(name='waldur_core.users.send_invitation_created')
def send_invitation_created(invitation_uuid, sender):
    """
    Invitation notification is sent to the user so that he can accept it and receive permissions.
    """
    invitation = models.Invitation.objects.get(uuid=invitation_uuid)
    context = utils.get_invitation_context(invitation, sender)
    context['link'] = invitation.link_template.format(uuid=invitation_uuid)

    logger.debug('About to send invitation to {email} to join {name} {type} as {role}'.format(
        email=invitation.email, **context))
    try:
        broadcast_mail('users', 'invitation_created', context, [invitation.email])
    except SMTPException as e:
        invitation.error_message = str(e)
        invitation.save(update_fields=['error_message'])
        raise


@shared_task(name='waldur_core.users.send_invitation_requested')
def send_invitation_requested(invitation_uuid, sender):
    """
    Invitation request is sent to staff users so that they can approve or reject invitation.
    """
    invitation = models.Invitation.objects.get(uuid=invitation_uuid)
    base_context = utils.get_invitation_context(invitation, sender)

    staff_users = core_models.User.objects.filter(is_staff=True, is_active=True).exclude(email='')
    for user in staff_users:
        token = utils.get_invitation_token(invitation, user)
        approve_link = settings.WALDUR_CORE['INVITATION_APPROVE_URL'].format(token=token)
        reject_link = settings.WALDUR_CORE['INVITATION_REJECT_URL'].format(token=token)
        context = dict(approve_link=approve_link, reject_link=reject_link, **base_context)
        broadcast_mail('users', 'invitation_requested', context, [user.email])


@shared_task(name='waldur_core.users.send_invitation_rejected')
def send_invitation_rejected(invitation_uuid, sender):
    """
    Invitation notification is sent to the user which has created invitation.
    """
    invitation = models.Invitation.objects.get(uuid=invitation_uuid)
    context = utils.get_invitation_context(invitation, sender)
    broadcast_mail('users', 'invitation_rejected', context, [invitation.created_by.email])


@shared_task(name='waldur_core.users.get_or_create_user')
def get_or_create_user(invitation_uuid, sender):
    invitation = models.Invitation.objects.get(uuid=invitation_uuid)

    user, created = utils.get_or_create_user(invitation)
    username = generate_safe_username(user.username)
    password = pwgen()
    try:
        profile, created = utils.get_or_create_profile(user, username, password)
    except (freeipa_exceptions.FreeIPAError, requests.RequestException) as e:
        logger.exception('Unable to create FreeIPA profile for user with ID: %s', user.id)
        invitation.error_message = str(e)
        invitation.save(update_fields=['error_message'])
        raise
    if created:
        sender = invitation.created_by.full_name or invitation.created_by.username
        context = utils.get_invitation_context(invitation, sender)
        context['username'] = username
        context['password'] = password
        context['link'] = invitation.link_template.format(uuid=invitation_uuid)
        broadcast_mail('users', 'invitation_approved', context, [invitation.email])
    else:
        send_invitation_created(invitation_uuid, sender)


@shared_task(name='waldur_core.users.process_invitation')
def process_invitation(invitation_uuid, sender):
    if settings.WALDUR_CORE['INVITATION_CREATE_MISSING_USER']:
        get_or_create_user(invitation_uuid, sender)
    else:
        send_invitation_created(invitation_uuid, sender)



from __future__ import unicode_literals

from django.core.management.base import BaseCommand
from django.db import transaction

from waldur_core.quotas import models, fields, exceptions, signals
from waldur_core.quotas.utils import get_models_with_quotas


class Command(BaseCommand):
    """ Recalculate all quotas """

    def handle(self, *args, **options):
        # TODO: implement other quotas recalculation
        # TODO: implement global stale quotas deletion
        self.delete_stale_quotas()
        self.init_missing_quotas()
        self.recalculate_global_quotas()
        self.recalculate_counter_quotas()
        self.recalculate_aggregator_quotas()
        self.stdout.write('XXX: Second time to make sure that aggregators of aggregators where calculated properly.')
        self.recalculate_aggregator_quotas()
        self.recalculate_custom_quotas()

    def delete_stale_quotas(self):
        self.stdout.write('Deleting stale quotas')
        for model in get_models_with_quotas():
            for obj in model.objects.all():
                quotas_names = model.get_quotas_names()
                obj.quotas.exclude(name__in=quotas_names).delete()
        self.stdout.write('...done')

    def init_missing_quotas(self):
        self.stdout.write('Initializing missing quotas')
        for model in get_models_with_quotas():
            for obj in model.objects.all():
                for field in obj.get_quotas_fields():
                    try:
                        field.get_or_create_quota(scope=obj)
                    except exceptions.CreationConditionFailedQuotaError:
                        pass
        self.stdout.write('...done')

    def recalculate_global_quotas(self):
        self.stdout.write('Recalculating global quotas')
        for model in get_models_with_quotas():
            if hasattr(model, 'GLOBAL_COUNT_QUOTA_NAME'):
                with transaction.atomic():
                    quota, _ = models.Quota.objects.get_or_create(name=model.GLOBAL_COUNT_QUOTA_NAME)
                    quota.usage = model.objects.count()
                    quota.save()
        self.stdout.write('...done')

    def recalculate_counter_quotas(self):
        self.stdout.write('Recalculating counter quotas')
        for model in get_models_with_quotas():
            for counter_field in model.get_quotas_fields(field_class=fields.CounterQuotaField):
                for instance in model.objects.all():
                    counter_field.recalculate(scope=instance)
        self.stdout.write('...done')

    def recalculate_aggregator_quotas(self):
        # TODO: recalculate child quotas first
        self.stdout.write('Recalculating aggregator quotas')
        for model in get_models_with_quotas():
            for aggregator_field in model.get_quotas_fields(field_class=fields.AggregatorQuotaField):
                for instance in model.objects.all():
                    aggregator_field.recalculate(scope=instance)
        self.stdout.write('...done')

    def recalculate_custom_quotas(self):
        self.stdout.write('Recalculating custom quotas')
        signals.recalculate_quotas.send(sender=self)
        self.stdout.write('...done')



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-01-03 13:39
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('waldur_jira', '0008_unique_props'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='project',
            name='available_for_all',
        ),
        migrations.RemoveField(
            model_name='project',
            name='reporter_field',
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-04-19 15:03
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import waldur_core.core.fields
import waldur_core.core.models
import waldur_core.core.validators


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0001_squashed_0054'),
        ('waldur_rijkscloud', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='FloatingIP',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('backend_id', models.CharField(db_index=True, max_length=255)),
                ('address', models.GenericIPAddressField(null=True, protocol='IPv4')),
                ('is_available', models.BooleanField(default=True)),
                ('settings', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='structure.ServiceSettings')),
            ],
            options={
                'verbose_name': 'Floating IP',
                'verbose_name_plural': 'Floating IPs',
            },
            bases=(waldur_core.core.models.BackendModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name='InternalIP',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('backend_id', models.CharField(db_index=True, max_length=255)),
                ('address', models.GenericIPAddressField(protocol='IPv4')),
                ('is_available', models.BooleanField(default=True)),
                ('settings', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='structure.ServiceSettings')),
            ],
            options={
                'verbose_name': 'Internal IP',
                'verbose_name_plural': 'Internal IPs',
            },
            bases=(waldur_core.core.models.BackendModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name='Network',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('backend_id', models.CharField(db_index=True, max_length=255)),
                ('settings', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='structure.ServiceSettings')),
            ],
            options={
                'abstract': False,
            },
            bases=(waldur_core.core.models.BackendModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name='SubNet',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('backend_id', models.CharField(db_index=True, max_length=255)),
                ('cidr', models.CharField(max_length=32)),
                ('gateway_ip', models.GenericIPAddressField(protocol='IPv4')),
                ('allocation_pools', waldur_core.core.fields.JSONField()),
                ('dns_nameservers', waldur_core.core.fields.JSONField(help_text='List of DNS name servers associated with the subnet.')),
                ('network', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='subnets', to='waldur_rijkscloud.Network')),
                ('settings', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='structure.ServiceSettings')),
            ],
            options={
                'verbose_name': 'Subnet',
                'verbose_name_plural': 'Subnets',
            },
            bases=(waldur_core.core.models.BackendModelMixin, models.Model),
        ),
        migrations.AddField(
            model_name='internalip',
            name='subnet',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='internal_ips', to='waldur_rijkscloud.SubNet'),
        ),
        migrations.AddField(
            model_name='instance',
            name='floating_ip',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='waldur_rijkscloud.FloatingIP'),
        ),
        migrations.AddField(
            model_name='instance',
            name='internal_ip',
            field=models.ForeignKey(default=None, on_delete=django.db.models.deletion.CASCADE, to='waldur_rijkscloud.InternalIP'),
            preserve_default=False,
        ),
        migrations.AlterUniqueTogether(
            name='subnet',
            unique_together=set([('settings', 'backend_id')]),
        ),
        migrations.AlterUniqueTogether(
            name='network',
            unique_together=set([('settings', 'backend_id')]),
        ),
        migrations.AlterUniqueTogether(
            name='internalip',
            unique_together=set([('settings', 'backend_id')]),
        ),
        migrations.AlterUniqueTogether(
            name='floatingip',
            unique_together=set([('settings', 'backend_id')]),
        ),
    ]



import logging

from django_filters.rest_framework import DjangoFilterBackend
from rest_framework.mixins import ListModelMixin
from rest_framework.viewsets import GenericViewSet

from waldur_core.structure import views as structure_views, filters as structure_filters

from . import filters, managers, models, serializers

logger = logging.getLogger(__name__)


def get_applications_queryset():
    return managers.ApplicationSummaryQuerySet(models.ApplicationModel.get_application_models())


def get_project_apps_count(project):
    return get_applications_queryset().filter(project=project).count()


structure_views.ProjectCountersView.register_counter('ansible', get_project_apps_count)


class ApplicationsSummaryViewSet(ListModelMixin, GenericViewSet):
    serializer_class = serializers.SummaryApplicationSerializer
    filter_backends = (structure_filters.GenericRoleFilter, DjangoFilterBackend)
    filter_class = filters.ApplicationFilter

    def get_queryset(self):
        return get_applications_queryset()



# -*- coding: utf-8 -*-
# Generated by Django 1.11.18 on 2019-02-26 09:10
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('support', '0034_bootstrap_statuses'),
    ]

    operations = [
        migrations.AddField(
            model_name='offering',
            name='plan',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='support.OfferingPlan'),
        ),
    ]



from __future__ import unicode_literals

from celery import chain
from django.db import transaction

from waldur_core.core import executors as core_executors
from waldur_core.core import tasks as core_tasks
from waldur_core.core import utils as core_utils
from waldur_core.structure import executors as structure_executors
from waldur_openstack.openstack import executors as openstack_executors
from waldur_openstack.openstack import models as openstack_models

from . import tasks, models


class VolumeCreateExecutor(core_executors.CreateExecutor):

    @classmethod
    def get_task_signature(cls, volume, serialized_volume, **kwargs):
        return chain(
            tasks.ThrottleProvisionTask().si(
                serialized_volume,
                'create_volume',
                state_transition='begin_creating'
            ),
            core_tasks.PollRuntimeStateTask().si(
                serialized_volume,
                backend_pull_method='pull_volume_runtime_state',
                success_state='available',
                erred_state='error',
            ).set(countdown=30)
        )


class VolumeUpdateExecutor(core_executors.UpdateExecutor):

    @classmethod
    def get_task_signature(cls, volume, serialized_volume, **kwargs):
        updated_fields = kwargs['updated_fields']
        if 'name' in updated_fields or 'description' in updated_fields:
            return core_tasks.BackendMethodTask().si(
                serialized_volume, 'update_volume', state_transition='begin_updating')
        else:
            return core_tasks.StateTransitionTask().si(serialized_volume, state_transition='begin_updating')


class VolumeDeleteExecutor(core_executors.DeleteExecutor):

    @classmethod
    def get_task_signature(cls, volume, serialized_volume, **kwargs):
        if volume.backend_id:
            return chain(
                core_tasks.BackendMethodTask().si(
                    serialized_volume, 'delete_volume', state_transition='begin_deleting'),
                core_tasks.PollBackendCheckTask().si(serialized_volume, 'is_volume_deleted'),
            )
        else:
            return core_tasks.StateTransitionTask().si(serialized_volume, state_transition='begin_deleting')


class VolumePullExecutor(core_executors.ActionExecutor):
    action = 'Pull'

    @classmethod
    def get_task_signature(cls, volume, serialized_volume, **kwargs):
        return core_tasks.BackendMethodTask().si(
            serialized_volume, 'pull_volume',
            state_transition='begin_updating')


class VolumeExtendExecutor(core_executors.ActionExecutor):
    action = 'Extend'

    @classmethod
    def get_action_details(cls, volume, **kwargs):
        return {
            'message': 'Extend volume from %s MB to %s MB' % (kwargs.get('old_size'), volume.size),
            'old_size': kwargs.get('old_size'),
            'new_size': volume.size,
        }

    @classmethod
    def pre_apply(cls, volume, **kwargs):
        super(VolumeExtendExecutor, cls).pre_apply(volume, **kwargs)
        if volume.instance is not None:
            volume.instance.action = 'Extend volume'
            volume.instance.schedule_updating()
            volume.instance.save()

    @classmethod
    def get_task_signature(cls, volume, serialized_volume, **kwargs):
        if volume.instance is None:
            return chain(
                core_tasks.BackendMethodTask().si(
                    serialized_volume,
                    backend_method='extend_volume',
                    state_transition='begin_updating',
                ),
                core_tasks.PollRuntimeStateTask().si(
                    serialized_volume,
                    backend_pull_method='pull_volume_runtime_state',
                    success_state='available',
                    erred_state='error'
                )
            )

        return chain(
            core_tasks.StateTransitionTask().si(
                core_utils.serialize_instance(volume.instance),
                state_transition='begin_updating'
            ),
            core_tasks.BackendMethodTask().si(
                serialized_volume,
                backend_method='detach_volume',
                state_transition='begin_updating'
            ),
            core_tasks.PollRuntimeStateTask().si(
                serialized_volume,
                backend_pull_method='pull_volume_runtime_state',
                success_state='available',
                erred_state='error'
            ),
            core_tasks.BackendMethodTask().si(
                serialized_volume,
                backend_method='extend_volume',
            ),
            core_tasks.PollRuntimeStateTask().si(
                serialized_volume,
                backend_pull_method='pull_volume_runtime_state',
                success_state='available',
                erred_state='error'
            ),
            core_tasks.BackendMethodTask().si(
                serialized_volume,
                instance_uuid=volume.instance.uuid.hex,
                device=volume.device,
                backend_method='attach_volume',
            ),
            core_tasks.PollRuntimeStateTask().si(
                serialized_volume,
                backend_pull_method='pull_volume_runtime_state',
                success_state='in-use',
                erred_state='error'
            ),
        )

    @classmethod
    def get_success_signature(cls, volume, serialized_volume, **kwargs):
        if volume.instance is None:
            return super(VolumeExtendExecutor, cls).get_success_signature(volume, serialized_volume, **kwargs)
        else:
            instance = volume.instance
            serialized_instance = core_utils.serialize_instance(instance)
            return chain(
                super(VolumeExtendExecutor, cls).get_success_signature(volume, serialized_volume, **kwargs),
                super(VolumeExtendExecutor, cls).get_success_signature(instance, serialized_instance, **kwargs),
            )

    @classmethod
    def get_failure_signature(cls, volume, serialized_volume, **kwargs):
        return tasks.VolumeExtendErredTask().s(serialized_volume)


class VolumeAttachExecutor(core_executors.ActionExecutor):
    action = 'Attach'

    @classmethod
    def get_action_details(cls, volume, **kwargs):
        return {'message': 'Attach volume to instance %s' % volume.instance.name}

    @classmethod
    def get_task_signature(cls, volume, serialized_volume, **kwargs):
        return chain(
            core_tasks.BackendMethodTask().si(
                serialized_volume,
                instance_uuid=volume.instance.uuid.hex,
                device=volume.device,
                backend_method='attach_volume',
                state_transition='begin_updating'
            ),
            core_tasks.PollRuntimeStateTask().si(
                serialized_volume,
                backend_pull_method='pull_volume_runtime_state',
                success_state='in-use',
                erred_state='error',
            ),
            # additional pull to populate field "device".
            core_tasks.BackendMethodTask().si(serialized_volume, backend_method='pull_volume'),
        )


class VolumeDetachExecutor(core_executors.ActionExecutor):
    action = 'Detach'

    @classmethod
    def get_action_details(cls, volume, **kwargs):
        return {'message': 'Detach volume from instance %s' % volume.instance.name}

    @classmethod
    def get_task_signature(cls, volume, serialized_volume, **kwargs):
        return chain(
            core_tasks.BackendMethodTask().si(
                serialized_volume, backend_method='detach_volume', state_transition='begin_updating'),
            core_tasks.PollRuntimeStateTask().si(
                serialized_volume,
                backend_pull_method='pull_volume_runtime_state',
                success_state='available',
                erred_state='error',
            )
        )


class SnapshotCreateExecutor(core_executors.CreateExecutor):

    @classmethod
    def get_task_signature(cls, snapshot, serialized_snapshot, **kwargs):
        return chain(
            tasks.ThrottleProvisionTask().si(
                serialized_snapshot,
                'create_snapshot',
                state_transition='begin_creating'
            ),
            core_tasks.PollRuntimeStateTask().si(
                serialized_snapshot,
                backend_pull_method='pull_snapshot_runtime_state',
                success_state='available',
                erred_state='error',
            ).set(countdown=10)
        )


class SnapshotUpdateExecutor(core_executors.UpdateExecutor):

    @classmethod
    def get_task_signature(cls, snapshot, serialized_snapshot, **kwargs):
        updated_fields = kwargs['updated_fields']
        # TODO: call separate task on metadata update
        if 'name' in updated_fields or 'description' in updated_fields:
            return core_tasks.BackendMethodTask().si(
                serialized_snapshot, 'update_snapshot', state_transition='begin_updating')
        else:
            return core_tasks.StateTransitionTask().si(serialized_snapshot, state_transition='begin_updating')


class SnapshotDeleteExecutor(core_executors.DeleteExecutor):

    @classmethod
    def get_task_signature(cls, snapshot, serialized_snapshot, **kwargs):
        if snapshot.backend_id:
            return chain(
                core_tasks.BackendMethodTask().si(
                    serialized_snapshot, 'delete_snapshot', state_transition='begin_deleting'),
                core_tasks.PollBackendCheckTask().si(serialized_snapshot, 'is_snapshot_deleted'),
            )
        else:
            return core_tasks.StateTransitionTask().si(serialized_snapshot, state_transition='begin_deleting')


class SnapshotPullExecutor(core_executors.ActionExecutor):
    action = 'Pull'

    @classmethod
    def get_task_signature(cls, snapshot, serialized_snapshot, **kwargs):
        return core_tasks.BackendMethodTask().si(
            serialized_snapshot, 'pull_snapshot',
            state_transition='begin_updating')


class InstanceCreateExecutor(core_executors.CreateExecutor):

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, ssh_key=None, flavor=None):
        serialized_volumes = [core_utils.serialize_instance(volume) for volume in instance.volumes.all()]
        _tasks = [tasks.ThrottleProvisionStateTask().si(serialized_instance, state_transition='begin_creating')]
        _tasks += cls.create_volumes(serialized_volumes)
        _tasks += cls.create_internal_ips(serialized_instance)
        _tasks += cls.create_instance(serialized_instance, flavor, ssh_key)
        _tasks += cls.pull_volumes(serialized_volumes)
        _tasks += cls.pull_security_groups(serialized_instance)
        _tasks += cls.create_floating_ips(instance, serialized_instance)
        return chain(*_tasks)

    @classmethod
    def create_volumes(cls, serialized_volumes):
        """
        Create all instance volumes and wait for them to provision.
        """
        _tasks = []

        # Create volumes
        for serialized_volume in serialized_volumes:
            _tasks.append(tasks.ThrottleProvisionTask().si(
                serialized_volume, 'create_volume', state_transition='begin_creating'))

        for index, serialized_volume in enumerate(serialized_volumes):
            # Wait for volume creation
            _tasks.append(core_tasks.PollRuntimeStateTask().si(
                serialized_volume,
                backend_pull_method='pull_volume_runtime_state',
                success_state='available',
                erred_state='error',
            ).set(countdown=30 if index == 0 else 0))

            # Pull volume runtime state
            _tasks.append(core_tasks.BackendMethodTask().si(
                serialized_volume,
                'pull_volume',
                update_fields=['runtime_state', 'bootable']
            ))

            # Mark volume as OK
            _tasks.append(core_tasks.StateTransitionTask().si(serialized_volume, state_transition='set_ok'))

        return _tasks

    @classmethod
    def create_internal_ips(cls, serialized_instance):
        """
        Create all network ports for an OpenStack instance.
        Although OpenStack Nova REST API allows to create network ports implicitly,
        we're not using it, because it does not take into account subnets.
        See also: https://specs.openstack.org/openstack/nova-specs/specs/juno/approved/selecting-subnet-when-creating-vm.html
        Therefore we're creating network ports beforehand with correct subnet.
        """
        return [core_tasks.BackendMethodTask().si(serialized_instance, 'create_instance_internal_ips')]

    @classmethod
    def create_instance(cls, serialized_instance, flavor, ssh_key=None):
        """
        It is assumed that volumes and network ports have been created beforehand.
        """
        _tasks = []
        kwargs = {
            'backend_flavor_id': flavor.backend_id,
        }
        if ssh_key is not None:
            kwargs['public_key'] = ssh_key.public_key

        # Wait 10 seconds after volume creation due to OpenStack restrictions.
        _tasks.append(core_tasks.BackendMethodTask().si(
            serialized_instance, 'create_instance', **kwargs).set(countdown=10))

        # Wait for instance creation
        _tasks.append(core_tasks.PollRuntimeStateTask().si(
            serialized_instance,
            backend_pull_method='pull_instance_runtime_state',
            success_state=models.Instance.RuntimeStates.ACTIVE,
            erred_state=models.Instance.RuntimeStates.ERROR,
        ))
        return _tasks

    @classmethod
    def pull_volumes(cls, serialized_volumes):
        """
        Update volumes runtime state and device name
        """
        _tasks = []
        for serialized_volume in serialized_volumes:
            _tasks.append(core_tasks.BackendMethodTask().si(
                serialized_volume,
                backend_method='pull_volume',
                update_fields=['runtime_state', 'device']
            ))
        return _tasks

    @classmethod
    def pull_security_groups(cls, serialized_instance):
        return [core_tasks.BackendMethodTask().si(serialized_instance, 'pull_instance_security_groups')]

    @classmethod
    def create_floating_ips(cls, instance, serialized_instance):
        _tasks = []

        if not instance.floating_ips.exists():
            return _tasks

        # Create non-existing floating IPs
        for floating_ip in instance.floating_ips.filter(backend_id=''):
            serialized_floating_ip = core_utils.serialize_instance(floating_ip)
            _tasks.append(core_tasks.BackendMethodTask().si(serialized_floating_ip, 'create_floating_ip'))

        # Push instance floating IPs
        _tasks.append(core_tasks.BackendMethodTask().si(serialized_instance, 'push_instance_floating_ips'))

        # Wait for operation completion
        for index, floating_ip in enumerate(instance.floating_ips):
            _tasks.append(core_tasks.PollRuntimeStateTask().si(
                core_utils.serialize_instance(floating_ip),
                backend_pull_method='pull_floating_ip_runtime_state',
                success_state='ACTIVE',
                erred_state='ERRED',
            ).set(countdown=5 if not index else 0))

        shared_tenant = instance.service_project_link.service.settings.scope
        if shared_tenant:
            _tasks.append(openstack_executors.TenantPullFloatingIPsExecutor.as_signature(shared_tenant))

        return _tasks

    @classmethod
    def get_success_signature(cls, instance, serialized_instance, **kwargs):
        return tasks.SetInstanceOKTask().si(serialized_instance)

    @classmethod
    def get_failure_signature(cls, instance, serialized_instance, **kwargs):
        return tasks.SetInstanceErredTask().s(serialized_instance)


class InstanceUpdateExecutor(core_executors.UpdateExecutor):

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        updated_fields = kwargs['updated_fields']
        if 'name' in updated_fields:
            return core_tasks.BackendMethodTask().si(
                serialized_instance, 'update_instance', state_transition='begin_updating')
        else:
            return core_tasks.StateTransitionTask().si(serialized_instance, state_transition='begin_updating')


class InstanceUpdateSecurityGroupsExecutor(core_executors.ActionExecutor):
    action = 'Update security groups'

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        return core_tasks.BackendMethodTask().si(
            serialized_instance,
            backend_method='push_instance_security_groups',
            state_transition='begin_updating',
        )


class InstanceDeleteExecutor(core_executors.DeleteExecutor):

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, force=False, **kwargs):
        delete_volumes = kwargs.pop('delete_volumes', True)
        release_floating_ips = kwargs.pop('release_floating_ips', True)

        delete_instance_tasks = cls.get_delete_instance_tasks(serialized_instance)
        release_floating_ips_tasks = cls.get_release_floating_ips_tasks(instance, release_floating_ips)
        detach_volumes_tasks = cls.get_detach_data_volumes_tasks(instance)
        delete_volumes_tasks = cls.get_delete_data_volumes_tasks(instance)
        delete_internal_ips_tasks = cls.get_delete_internal_ips_tasks(serialized_instance)

        # Case 1. Instance does not exist at backend
        if not instance.backend_id:
            return chain(cls.get_delete_incomplete_instance_tasks(instance, serialized_instance))

        # Case 2. Instance exists at backend.
        # Data volumes are detached and deleted explicitly
        # because once volume is attached after instance is created,
        # it is not removed automatically.
        # System volume is deleted implicitly since delete_on_termination=True
        elif delete_volumes:
            return chain(
                detach_volumes_tasks +
                delete_volumes_tasks +
                delete_instance_tasks +
                release_floating_ips_tasks +
                delete_internal_ips_tasks
            )

        # Case 3. Instance exists at backend.
        # Data volumes are detached and not deleted.
        else:
            return chain(
                detach_volumes_tasks +
                delete_instance_tasks +
                release_floating_ips_tasks +
                delete_internal_ips_tasks
            )

    @classmethod
    def get_delete_incomplete_instance_tasks(cls, instance, serialized_instance):
        _tasks = []

        _tasks.append(core_tasks.StateTransitionTask().si(
            serialized_instance,
            state_transition='begin_deleting'
        ))

        _tasks += cls.get_delete_internal_ips_tasks(serialized_instance)

        for volume in instance.volumes.all():
            if volume.backend_id:
                serialized_volume = core_utils.serialize_instance(volume)
                _tasks.append(core_tasks.BackendMethodTask().si(
                    serialized_volume,
                    'delete_volume',
                    state_transition='begin_deleting'
                ))
                _tasks.append(core_tasks.PollBackendCheckTask().si(
                    serialized_volume,
                    'is_volume_deleted'
                ))

        _tasks += [tasks.DeleteIncompleteInstanceTask().si(serialized_instance)]

        return _tasks

    @classmethod
    def get_delete_internal_ips_tasks(cls, serialized_instance):
        """
        OpenStack Neutron ports should be deleted explicitly because we're creating them explicitly.
        Otherwise when port quota is exhausted, user is not able to provision new VMs anymore.
        """
        return [
            core_tasks.BackendMethodTask().si(
                serialized_instance,
                backend_method='delete_instance_internal_ips',
            )
        ]

    @classmethod
    def get_delete_instance_tasks(cls, serialized_instance):
        return [
            core_tasks.BackendMethodTask().si(
                serialized_instance,
                backend_method='delete_instance',
                state_transition='begin_deleting',
            ),
            core_tasks.PollBackendCheckTask().si(
                serialized_instance,
                backend_check_method='is_instance_deleted',
            ),
        ]

    @classmethod
    def get_release_floating_ips_tasks(cls, instance, release_floating_ips):
        _tasks = []
        if release_floating_ips:
            for index, floating_ip in enumerate(instance.floating_ips):
                _tasks.append(core_tasks.BackendMethodTask().si(
                    core_utils.serialize_instance(floating_ip),
                    'delete_floating_ip',
                ).set(countdown=5 if not index else 0))
        else:
            # pull related floating IPs state after instance deletion
            for index, floating_ip in enumerate(instance.floating_ips):
                _tasks.append(core_tasks.BackendMethodTask().si(
                    core_utils.serialize_instance(floating_ip),
                    'pull_floating_ip_runtime_state',
                ).set(countdown=5 if not index else 0))

        shared_tenant = instance.service_project_link.service.settings.scope
        if shared_tenant and isinstance(shared_tenant, openstack_models.Tenant):
            if shared_tenant.state == openstack_models.Tenant.States.OK:
                _tasks.append(openstack_executors.TenantPullFloatingIPsExecutor.as_signature(shared_tenant))

        return _tasks

    @classmethod
    def get_detach_data_volumes_tasks(cls, instance):
        data_volumes = instance.volumes.all().filter(bootable=False)
        detach_volumes = [
            core_tasks.BackendMethodTask().si(
                core_utils.serialize_instance(volume),
                backend_method='detach_volume',
            )
            for volume in data_volumes
        ]
        check_volumes = [
            core_tasks.PollRuntimeStateTask().si(
                core_utils.serialize_instance(volume),
                backend_pull_method='pull_volume_runtime_state',
                success_state='available',
                erred_state='error',
                deleted_state='deleted',
            )
            for volume in data_volumes
        ]
        return detach_volumes + check_volumes

    @classmethod
    def get_delete_data_volumes_tasks(cls, instance):
        data_volumes = instance.volumes.all().filter(bootable=False)
        return [
            VolumeDeleteExecutor.as_signature(volume)
            for volume in data_volumes
        ]


class InstanceFlavorChangeExecutor(core_executors.ActionExecutor):
    action = 'Change flavor'

    @classmethod
    def get_action_details(cls, instance, **kwargs):
        old_flavor_name = kwargs.get('old_flavor_name')
        new_flavor_name = kwargs.get('flavor').name
        return {
            'message': 'Change flavor from %s to %s' % (old_flavor_name, new_flavor_name),
            'old_flavor_name': old_flavor_name,
            'new_flavor_name': new_flavor_name,
        }

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        flavor = kwargs.pop('flavor')
        return chain(
            core_tasks.BackendMethodTask().si(
                serialized_instance,
                backend_method='resize_instance',
                state_transition='begin_updating',
                flavor_id=flavor.backend_id
            ),
            core_tasks.PollRuntimeStateTask().si(
                serialized_instance,
                backend_pull_method='pull_instance_runtime_state',
                success_state='VERIFY_RESIZE',
                erred_state='ERRED'
            ),
            core_tasks.BackendMethodTask().si(
                serialized_instance,
                backend_method='confirm_instance_resize'
            ),
            core_tasks.PollRuntimeStateTask().si(
                serialized_instance,
                backend_pull_method='pull_instance_runtime_state',
                success_state='SHUTOFF',
                erred_state='ERRED'
            ),
        )


class InstancePullExecutor(core_executors.ActionExecutor):
    action = 'Pull'

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        return chain(
            core_tasks.BackendMethodTask().si(
                serialized_instance, 'pull_instance', state_transition='begin_updating',
            ),
            core_tasks.BackendMethodTask().si(serialized_instance, 'pull_instance_security_groups'),
            core_tasks.BackendMethodTask().si(serialized_instance, 'pull_instance_internal_ips'),
            core_tasks.BackendMethodTask().si(serialized_instance, 'pull_instance_floating_ips'),
        )


class InstanceFloatingIPsUpdateExecutor(core_executors.ActionExecutor):
    action = 'Update floating IPs'

    @classmethod
    def get_action_details(cls, instance, **kwargs):
        attached = set(instance._new_floating_ips) - set(instance._old_floating_ips)
        detached = set(instance._old_floating_ips) - set(instance._new_floating_ips)

        messages = []
        if attached:
            messages.append('Attached floating IPs: %s.' % ', '.join(attached))

        if detached:
            messages.append('Detached floating IPs: %s.' % ', '.join(detached))

        if not messages:
            messages.append('Instance floating IPs have been updated.')

        return {
            'message': ' '.join(messages),
            'attached': list(attached),
            'detached': list(detached),
        }

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        _tasks = [core_tasks.StateTransitionTask().si(serialized_instance, state_transition='begin_updating')]
        # Create non-exist floating IPs
        for floating_ip in instance.floating_ips.filter(backend_id=''):
            serialized_floating_ip = core_utils.serialize_instance(floating_ip)
            _tasks.append(core_tasks.BackendMethodTask().si(serialized_floating_ip, 'create_floating_ip'))
        # Push instance floating IPs
        _tasks.append(core_tasks.BackendMethodTask().si(serialized_instance, 'push_instance_floating_ips'))
        # Wait for operation completion
        for index, floating_ip in enumerate(instance.floating_ips):
            _tasks.append(core_tasks.PollRuntimeStateTask().si(
                core_utils.serialize_instance(floating_ip),
                backend_pull_method='pull_floating_ip_runtime_state',
                success_state='ACTIVE',
                erred_state='ERRED',
            ).set(countdown=5 if not index else 0))
        # Pull floating IPs again to update state of disconnected IPs
        _tasks.append(core_tasks.IndependentBackendMethodTask().si(serialized_instance, 'pull_floating_ips'))
        return chain(*_tasks)

    @classmethod
    def get_success_signature(cls, instance, serialized_instance, **kwargs):
        return tasks.SetInstanceOKTask().si(serialized_instance)

    @classmethod
    def get_failure_signature(cls, instance, serialized_instance, **kwargs):
        return tasks.SetInstanceErredTask().s(serialized_instance)


class InstanceStopExecutor(core_executors.ActionExecutor):
    action = 'Stop'

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        return chain(
            core_tasks.BackendMethodTask().si(
                serialized_instance, 'stop_instance', state_transition='begin_updating',
            ),
            core_tasks.PollRuntimeStateTask().si(
                serialized_instance,
                backend_pull_method='pull_instance_runtime_state',
                success_state='SHUTOFF',
                erred_state='ERRED'
            ),
        )


class InstanceStartExecutor(core_executors.ActionExecutor):
    action = 'Start'

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        return chain(
            core_tasks.BackendMethodTask().si(
                serialized_instance, 'start_instance', state_transition='begin_updating',
            ),
            core_tasks.PollRuntimeStateTask().si(
                serialized_instance,
                backend_pull_method='pull_instance_runtime_state',
                success_state='ACTIVE',
                erred_state='ERRED'
            ),
        )


class InstanceRestartExecutor(core_executors.ActionExecutor):
    action = 'Restart'

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        return chain(
            core_tasks.BackendMethodTask().si(
                serialized_instance, 'restart_instance', state_transition='begin_updating',
            ),
            core_tasks.PollRuntimeStateTask().si(
                serialized_instance,
                backend_pull_method='pull_instance_runtime_state',
                success_state='ACTIVE',
                erred_state='ERRED'
            ),
        )


class InstanceInternalIPsSetUpdateExecutor(core_executors.ActionExecutor):
    action = 'Update internal IPs'

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        return core_tasks.BackendMethodTask().si(
            serialized_instance, 'push_instance_internal_ips', state_transition='begin_updating',
        )


class BackupCreateExecutor(core_executors.CreateExecutor):

    @classmethod
    def get_task_signature(cls, backup, serialized_backup, **kwargs):
        serialized_snapshots = [core_utils.serialize_instance(snapshot) for snapshot in backup.snapshots.all()]

        _tasks = [core_tasks.StateTransitionTask().si(serialized_backup, state_transition='begin_creating')]
        for serialized_snapshot in serialized_snapshots:
            _tasks.append(tasks.ThrottleProvisionTask().si(
                serialized_snapshot, 'create_snapshot', force=True, state_transition='begin_creating'))
        for index, serialized_snapshot in enumerate(serialized_snapshots):
            _tasks.append(core_tasks.PollRuntimeStateTask().si(
                serialized_snapshot,
                backend_pull_method='pull_snapshot_runtime_state',
                success_state='available',
                erred_state='error',
            ).set(countdown=10 if index == 0 else 0))
            _tasks.append(core_tasks.StateTransitionTask().si(serialized_snapshot, state_transition='set_ok'))

        return chain(*_tasks)

    @classmethod
    def get_failure_signature(cls, backup, serialized_backup, **kwargs):
        return tasks.SetBackupErredTask().s(serialized_backup)


class BackupDeleteExecutor(core_executors.DeleteExecutor):

    @classmethod
    @transaction.atomic
    def pre_apply(cls, backup, **kwargs):
        for snapshot in backup.snapshots.all():
            snapshot.schedule_deleting()
            snapshot.save(update_fields=['state'])
        core_executors.DeleteExecutor.pre_apply(backup)

    @classmethod
    def get_task_signature(cls, backup, serialized_backup, force=False, **kwargs):
        serialized_snapshots = [core_utils.serialize_instance(snapshot) for snapshot in backup.snapshots.all()]

        _tasks = [core_tasks.StateTransitionTask().si(serialized_backup, state_transition='begin_deleting')]
        for serialized_snapshot in serialized_snapshots:
            _tasks.append(core_tasks.BackendMethodTask().si(
                serialized_snapshot, 'delete_snapshot', state_transition='begin_deleting'))
        for serialized_snapshot in serialized_snapshots:
            _tasks.append(core_tasks.PollBackendCheckTask().si(serialized_snapshot, 'is_snapshot_deleted'))
            _tasks.append(core_tasks.DeletionTask().si(serialized_snapshot))

        return chain(*_tasks)

    @classmethod
    def get_failure_signature(cls, backup, serialized_backup, force=False, **kwargs):
        if not force:
            return tasks.SetBackupErredTask().s(serialized_backup)
        else:
            return tasks.ForceDeleteBackupTask().si(serialized_backup)


class SnapshotRestorationExecutor(core_executors.CreateExecutor):
    """ Restores volume from snapshot instance """

    @classmethod
    def get_task_signature(cls, snapshot_restoration, serialized_snapshot_restoration, **kwargs):
        serialized_volume = core_utils.serialize_instance(snapshot_restoration.volume)

        _tasks = [
            tasks.ThrottleProvisionTask().si(
                serialized_volume, 'create_volume', state_transition='begin_creating'),
            core_tasks.PollRuntimeStateTask().si(
                serialized_volume, 'pull_volume_runtime_state', success_state='available', erred_state='error',
            ).set(countdown=30),
            core_tasks.BackendMethodTask().si(serialized_volume, 'remove_bootable_flag'),
            core_tasks.BackendMethodTask().si(serialized_volume, 'pull_volume'),
        ]

        return chain(*_tasks)

    @classmethod
    def get_success_signature(cls, snapshot_restoration, serialized_snapshot_restoration, **kwargs):
        serialized_volume = core_utils.serialize_instance(snapshot_restoration.volume)
        return core_tasks.StateTransitionTask().si(serialized_volume, state_transition='set_ok')

    @classmethod
    def get_failure_signature(cls, snapshot_restoration, serialized_snapshot_restoration, **kwargs):
        serialized_volume = core_utils.serialize_instance(snapshot_restoration.volume)
        return core_tasks.StateTransitionTask().si(serialized_volume, state_transition='set_erred')


class OpenStackTenantCleanupExecutor(structure_executors.BaseCleanupExecutor):
    related_executor = openstack_executors.OpenStackCleanupExecutor

    pre_models = (
        models.SnapshotSchedule,
        models.BackupSchedule,
    )

    executors = (
        (models.Snapshot, SnapshotDeleteExecutor),
        (models.Backup, BackupDeleteExecutor),
        (models.Instance, InstanceDeleteExecutor),
        (models.Volume, VolumeDeleteExecutor),
    )



from __future__ import unicode_literals

from waldur_core.core import WaldurExtension


class BookingExtension(WaldurExtension):
    class Settings:
        pass

    @staticmethod
    def django_app():
        return 'waldur_mastermind.booking'

    @staticmethod
    def is_assembly():
        return True



# -*- coding: utf-8 -*-
# Generated by Django 1.11.16 on 2018-12-28 13:46
from __future__ import unicode_literals

from django.db import migrations, models
import django_fsm


class Migration(migrations.Migration):

    dependencies = [
        ('support', '0033_issuestatus'),
    ]

    def bootstrap_default_statuses(apps, schema_editor):
        IssueStatus = apps.get_model('support', 'IssueStatus')

        class Types(object):
            RESOLVED = 0
            CANCELED = 1

        for status in ['Resolved', 'Closed', 'Completed', 'Done']:
            IssueStatus.objects.create(
                name=status,
                type=Types.RESOLVED
            )

        for status in ['Canceled', 'Declined']:
            IssueStatus.objects.create(
                name=status,
                type=Types.CANCELED
            )

    operations = [
        migrations.RunPython(bootstrap_default_statuses),
    ]



from __future__ import unicode_literals

from django.utils import timezone

from waldur_core.structure import models as structure_models

from . import models, utils


def format_event(tags):
    return {
        'measurement': 'events',
        'tags': tags,
        # fields is required, but empty
        'fields': {
            'value': 0
        }
    }


def log_resource_created(sender, instance, **kwargs):
    client = utils.get_influxdb_client()
    if not client:
        return
    title = 'Resource {resource_name} has been created.'.format(resource_name=instance.full_name)
    point = format_event({
        'title': title,
        'type': 'resource_created',
    })
    utils.write_points(client, [point])


def log_resource_deleted(sender, instance, **kwargs):
    client = utils.get_influxdb_client()
    if not client:
        return
    title = 'Resource {resource_name} has been deleted.'.format(resource_name=instance.full_name)
    point = format_event({
        'title': title,
        'type': 'resource_deleted',
    })
    utils.write_points(client, [point])


def update_daily_quotas(sender, instance, created=False, **kwargs):
    if not isinstance(instance.scope, (structure_models.Project, structure_models.Customer)):
        return

    if not created and not instance.tracker.has_changed('usage'):
        return

    models.DailyQuotaHistory.objects.update_or_create_quota(
        scope=instance.scope,
        name=instance.name,
        date=timezone.now().date(),
        usage=instance.usage,
    )



import django.dispatch

# TODO: Make all the serializers emit this signal
pre_serializer_fields = django.dispatch.Signal(providing_args=['fields'])

# This signal allows to implement deletion validation in dependent
# application without introducing circular dependency
pre_delete_validate = django.dispatch.Signal(providing_args=['instance', 'user'])
